"""
    :codeauthor: Lukas Raska <lukas@raska.me>
"""

from salt.exceptions import CommandExecutionError, SaltInvocationError
from salt.modules import elasticsearch
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase, skipIf

# Import elasticsearch exceptions
NO_ELASTIC = False
try:
    from elasticsearch import TransportError, NotFoundError
except Exception:  # pylint: disable=broad-except
    NO_ELASTIC = True


@skipIf(NO_ELASTIC, "Install elasticsearch-py before running Elasticsearch unit tests.")
class ElasticsearchTestCase(TestCase):
    """
    Test cases for salt.modules.elasticsearch
    """

    @staticmethod
    def es_return_true(hosts=None, profile=None):
        return True

    @staticmethod
    def es_raise_command_execution_error(hosts=None, profile=None):
        raise CommandExecutionError("custom message")

    # 'ping' function tests: 2

    def test_ping(self):
        """
        Test if ping succeeds
        """
        with patch.object(elasticsearch, "_get_instance", self.es_return_true):
            self.assertTrue(elasticsearch.ping())

    def test_ping_failure(self):
        """
        Test if ping fails
        """
        with patch.object(
            elasticsearch, "_get_instance", self.es_raise_command_execution_error
        ):
            self.assertFalse(elasticsearch.ping())

    # 'info' function tests: 2

    def test_info(self):
        """
        Test if status fetch succeeds
        """

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            def info(self):
                """
                Mock of info method
                """
                return [{"test": "key"}]

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertListEqual(elasticsearch.info(), [{"test": "key"}])

    def test_info_failure(self):
        """
        Test if status fetch fails with CommandExecutionError
        """

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            def info(self):
                """
                Mock of info method
                """
                raise TransportError("custom error", 123)

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertRaises(CommandExecutionError, elasticsearch.info)

    # 'node_info' function tests: 2

    def test_node_info(self):
        """
        Test if node status fetch succeeds
        """

        class MockElasticNodes:
            """
            Mock of Elasticsearch NodesClient
            """

            def info(self, node_id=None, flat_settings=None):
                """
                Mock of info method
                """
                return [{"test": "key"}]

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            nodes = MockElasticNodes()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertListEqual(elasticsearch.node_info(), [{"test": "key"}])

    def test_node_info_failure(self):
        """
        Test if node status fetch fails with CommandExecutionError
        """

        class MockElasticNodes:
            """
            Mock of Elasticsearch NodesClient
            """

            def info(self, node_id=None, flat_settings=None):
                """
                Mock of info method
                """
                raise TransportError("custom error", 123)

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            nodes = MockElasticNodes()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertRaises(CommandExecutionError, elasticsearch.node_info)

    # 'cluster_health' function tests: 2

    def test_cluster_health(self):
        """
        Test if cluster status health fetch succeeds
        """

        class MockElasticCluster:
            """
            Mock of Elasticsearch ClusterClient
            """

            def health(self, index=None, level=None, local=None):
                """
                Mock of health method
                """
                return [{"test": "key"}]

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            cluster = MockElasticCluster()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertListEqual(elasticsearch.cluster_health(), [{"test": "key"}])

    def test_cluster_health_failure(self):
        """
        Test if cluster status health fetch fails with CommandExecutionError
        """

        class MockElasticCluster:
            """
            Mock of Elasticsearch ClusterClient
            """

            def health(self, index=None, level=None, local=None):
                """
                Mock of health method
                """
                raise TransportError("custom error", 123)

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            cluster = MockElasticCluster()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertRaises(CommandExecutionError, elasticsearch.cluster_health)

    # 'cluster_stats' function tests: 2

    def test_cluster_stats(self):
        """
        Test if cluster stats fetch succeeds
        """

        class MockElasticCluster:
            """
            Mock of Elasticsearch ClusterClient
            """

            def stats(self, node_id=None):
                """
                Mock of health method
                """
                return [{"test": "key"}]

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            cluster = MockElasticCluster()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertListEqual(elasticsearch.cluster_stats(), [{"test": "key"}])

    def test_cluster_stats_failure(self):
        """
        Test if cluster stats fetch fails with CommandExecutionError
        """

        class MockElasticCluster:
            """
            Mock of Elasticsearch ClusterClient
            """

            def stats(self, node_id=None):
                """
                Mock of health method
                """
                raise TransportError("custom error", 123)

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            cluster = MockElasticCluster()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertRaises(CommandExecutionError, elasticsearch.cluster_stats)

    # 'alias_create' function tests: 3

    def test_alias_create(self):
        """
        Test if alias is created
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def put_alias(self, index=None, name=None, body=None):
                """
                Mock of put_alias method
                """
                return {"acknowledged": True}

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertTrue(elasticsearch.alias_create("foo", "bar", body="baz"))

    def test_alias_create_unack(self):
        """
        Test if alias creation is not acked
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def put_alias(self, index=None, name=None, body=None):
                """
                Mock of put_alias method
                """
                return {"acknowledged": False}

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertFalse(elasticsearch.alias_create("foo", "bar", body="baz"))

    def test_alias_create_failure(self):
        """
        Test if alias creation fails with CommandExecutionError
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def put_alias(self, index=None, name=None, body=None):
                """
                Mock of put_alias method
                """
                raise TransportError("custom message", 123)

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertRaises(
                CommandExecutionError,
                elasticsearch.alias_create,
                "foo",
                "bar",
                body="baz",
            )

    # 'alias_delete' function tests: 3

    def test_alias_delete(self):
        """
        Test if alias is deleted
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def delete_alias(self, index=None, name=None):
                """
                Mock of delete_alias method
                """
                return {"acknowledged": True}

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertTrue(elasticsearch.alias_delete("foo", "bar"))

    def test_alias_delete_unack(self):
        """
        Test if alias deletion is not acked
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def delete_alias(self, index=None, name=None):
                """
                Mock of delete_alias method
                """
                return {"acknowledged": False}

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertFalse(elasticsearch.alias_delete("foo", "bar"))

    def test_alias_delete_failure(self):
        """
        Test if alias deletion fails with CommandExecutionError
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def delete_alias(self, index=None, name=None):
                """
                Mock of delete_alias method
                """
                raise TransportError("custom message", 123)

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertRaises(
                CommandExecutionError, elasticsearch.alias_delete, "foo", "bar"
            )

    # 'alias_exists' function tests: 3

    def test_alias_exists(self):
        """
        Test if alias exists
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def exists_alias(self, index=None, name=None):
                """
                Mock of exists_alias method
                """
                return {"acknowledged": True}

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertTrue(elasticsearch.alias_exists("foo", "bar"))

    def test_alias_exists_not(self):
        """
        Test if alias doesn't exist
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def exists_alias(self, index=None, name=None):
                """
                Mock of exists_alias method
                """
                raise NotFoundError

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertFalse(elasticsearch.alias_exists("foo", "bar"))

    def test_alias_exists_failure(self):
        """
        Test if alias status obtain fails with CommandExecutionError
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def exists_alias(self, index=None, name=None):
                """
                Mock of exists_alias method
                """
                raise TransportError("custom message", 123)

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertRaises(
                CommandExecutionError, elasticsearch.alias_exists, "foo", "bar"
            )

    # 'alias_get' function tests: 3

    def test_alias_get(self):
        """
        Test if alias can be obtained
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def get_alias(self, index=None, name=None):
                """
                Mock of get_alias method
                """
                return {"test": "key"}

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertDictEqual(elasticsearch.alias_get("foo", "bar"), {"test": "key"})

    def test_alias_get_not(self):
        """
        Test if alias doesn't exist
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def get_alias(self, index=None, name=None):
                """
                Mock of get_alias method
                """
                raise NotFoundError

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertIs(elasticsearch.alias_get("foo", "bar"), None)

    def test_alias_get_failure(self):
        """
        Test if alias obtain fails with CommandExecutionError
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def get_alias(self, index=None, name=None):
                """
                Mock of get_alias method
                """
                raise TransportError("custom message", 123)

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertRaises(
                CommandExecutionError, elasticsearch.alias_get, "foo", "bar"
            )

    # 'document_create' function tests: 2

    def test_document_create(self):
        """
        Test if document can be created
        """

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            def index(self, index=None, doc_type=None, body=None, id=None):
                """
                Mock of index method
                """
                return {"test": "key"}

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertDictEqual(
                elasticsearch.document_create("foo", "bar"), {"test": "key"}
            )

    def test_document_create_failure(self):
        """
        Test if document creation fails with CommandExecutionError
        """

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            def index(self, index=None, doc_type=None, body=None, id=None):
                """
                Mock of index method
                """
                raise TransportError("custom message", 123)

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertRaises(
                CommandExecutionError, elasticsearch.document_create, "foo", "bar"
            )

    # 'document_delete' function tests: 2

    def test_document_delete(self):
        """
        Test if document can be deleted
        """

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            def delete(self, index=None, doc_type=None, id=None):
                """
                Mock of index method
                """
                return {"test": "key"}

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertDictEqual(
                elasticsearch.document_delete("foo", "bar", "baz"), {"test": "key"}
            )

    def test_document_delete_failure(self):
        """
        Test if document deletion fails with CommandExecutionError
        """

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            def delete(self, index=None, doc_type=None, id=None):
                """
                Mock of index method
                """
                raise TransportError("custom message", 123)

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertRaises(
                CommandExecutionError,
                elasticsearch.document_delete,
                "foo",
                "bar",
                "baz",
            )

    # 'document_exists' function tests: 3

    def test_document_exists(self):
        """
        Test if document status can be obtained
        """

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            def exists(self, index=None, doc_type=None, id=None):
                """
                Mock of index method
                """
                return True

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertTrue(elasticsearch.document_exists("foo", "bar"))

    def test_document_exists_not(self):
        """
        Test if document doesn't exist
        """

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            def exists(self, index=None, doc_type=None, id=None):
                """
                Mock of index method
                """
                return False

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertFalse(elasticsearch.document_exists("foo", "bar"))

    def test_document_exists_failure(self):
        """
        Test if document exist state fails with CommandExecutionError
        """

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            def exists(self, index=None, doc_type=None, id=None):
                """
                Mock of index method
                """
                raise TransportError("custom message", 123)

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertRaises(
                CommandExecutionError, elasticsearch.document_exists, "foo", "bar"
            )

    # 'document_get' function tests: 3

    def test_document_get(self):
        """
        Test if document exists
        """

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            def get(self, index=None, doc_type=None, id=None):
                """
                Mock of index method
                """
                return {"test": "key"}

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertDictEqual(
                elasticsearch.document_get("foo", "bar"), {"test": "key"}
            )

    def test_document_get_not(self):
        """
        Test if document doesn't exit
        """

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            def get(self, index=None, doc_type=None, id=None):
                """
                Mock of index method
                """
                raise NotFoundError

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertIs(elasticsearch.document_get("foo", "bar"), None)

    def test_document_get_failure(self):
        """
        Test if document obtain fails with CommandExecutionError
        """

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            def get(self, index=None, doc_type=None, id=None):
                """
                Mock of index method
                """
                raise TransportError("custom message", 123)

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertRaises(
                CommandExecutionError, elasticsearch.document_get, "foo", "bar"
            )

    # 'index_create' function tests: 5

    def test_index_create(self):
        """
        Test if index can be created
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def create(self, index=None, body=None):
                """
                Mock of index method
                """
                return {"acknowledged": True, "shards_acknowledged": True}

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertTrue(elasticsearch.index_create("foo", "bar"))

    def test_index_create_no_shards(self):
        """
        Test if index is created and no shards info is returned
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def create(self, index=None, body=None):
                """
                Mock of index method
                """
                return {"acknowledged": True}

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertTrue(elasticsearch.index_create("foo", "bar"))

    def test_index_create_not_shards(self):
        """
        Test if index is created and shards didn't acked
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def create(self, index=None, body=None):
                """
                Mock of index method
                """
                return {"acknowledged": True, "shards_acknowledged": False}

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertFalse(elasticsearch.index_create("foo", "bar"))

    def test_index_create_not(self):
        """
        Test if index is created and shards didn't acked
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def create(self, index=None, body=None):
                """
                Mock of index method
                """
                return {"acknowledged": False, "shards_acknowledged": False}

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertFalse(elasticsearch.index_create("foo", "bar"))

    def test_index_create_failure(self):
        """
        Test if index creation fails with CommandExecutionError
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def create(self, index=None, body=None):
                """
                Mock of index method
                """
                raise TransportError("custom message", 123)

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertRaises(
                CommandExecutionError, elasticsearch.index_create, "foo", "bar"
            )

    # 'index_delete' function tests: 3

    def test_index_delete(self):
        """
        Test if index can be deleted
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def delete(self, index=None):
                """
                Mock of index method
                """
                return {"acknowledged": True}

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertTrue(elasticsearch.index_delete("foo", "bar"))

    def test_index_delete_not(self):
        """
        Test if index is deleted and shards didn't acked
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def delete(self, index=None):
                """
                Mock of index method
                """
                return {"acknowledged": False}

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertFalse(elasticsearch.index_delete("foo", "bar"))

    def test_index_delete_failure(self):
        """
        Test if index deletion fails with CommandExecutionError
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def delete(self, index=None):
                """
                Mock of index method
                """
                raise TransportError("custom message", 123)

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertRaises(
                CommandExecutionError, elasticsearch.index_delete, "foo", "bar"
            )

    # 'index_exists' function tests: 3

    def test_index_exists(self):
        """
        Test if index exists
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def exists(self, index=None):
                """
                Mock of index method
                """
                return True

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertTrue(elasticsearch.index_exists("foo", "bar"))

    def test_index_exists_not(self):
        """
        Test if index doesn't exist
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def exists(self, index=None):
                """
                Mock of index method
                """
                raise NotFoundError

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertFalse(elasticsearch.index_exists("foo", "bar"))

    def test_index_exists_failure(self):
        """
        Test if alias exist state fails with CommandExecutionError
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def exists(self, index=None):
                """
                Mock of index method
                """
                raise TransportError("custom message", 123)

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertRaises(
                CommandExecutionError, elasticsearch.index_exists, "foo", "bar"
            )

    def test_index_get_settings(self):
        """
        Test if settings can be obtained from the index
        """

        fake_es = MagicMock()
        fake_es.indices = MagicMock()
        fake_es.indices.get_settings = MagicMock(return_value={"foo": "key"})
        fake_instance = MagicMock(return_value=fake_es)

        with patch.object(elasticsearch, "_get_instance", fake_instance):
            self.assertDictEqual(
                elasticsearch.index_get_settings("foo", "bar"), {"foo": "key"}
            )

    def test_index_get_settings_not_exists(self):
        """
        Test index_get_settings if index doesn't exist
        """

        fake_es = MagicMock()
        fake_es.indices = MagicMock()
        fake_es.indices.get_settings = MagicMock()
        fake_es.indices.get_settings.side_effect = NotFoundError("custom error", 123)
        fake_instance = MagicMock(return_value=fake_es)

        with patch.object(elasticsearch, "_get_instance", fake_instance):
            self.assertIs(elasticsearch.index_get_settings(index="foo"), None)

    def test_get_settings_failure(self):
        """
        Test if index settings get fails with CommandExecutionError
        """

        fake_es = MagicMock()
        fake_es.indices = MagicMock()
        fake_es.indices.get_settings = MagicMock()
        fake_es.indices.get_settings.side_effect = TransportError("custom error", 123)
        fake_instance = MagicMock(return_value=fake_es)

        with patch.object(elasticsearch, "_get_instance", fake_instance):
            self.assertRaises(
                CommandExecutionError, elasticsearch.index_get_settings, index="foo"
            )

    def test_index_put_settings(self):
        """
        Test if we can put settings for the index
        """

        body = {"settings": {"index": {"number_of_replicas": 2}}}
        fake_es = MagicMock()
        fake_es.indices = MagicMock()
        fake_es.indices.put_settings = MagicMock(return_value={"acknowledged": True})
        fake_instance = MagicMock(return_value=fake_es)

        with patch.object(elasticsearch, "_get_instance", fake_instance):
            self.assertTrue(elasticsearch.index_put_settings(index="foo", body=body))

    def test_index_put_settings_not_exists(self):
        """
        Test if settings put executed against non-existing index
        """

        body = {"settings": {"index": {"number_of_replicas": 2}}}
        fake_es = MagicMock()
        fake_es.indices = MagicMock()
        fake_es.indices.put_settings = MagicMock()
        fake_es.indices.put_settings.side_effect = NotFoundError("custom error", 123)
        fake_instance = MagicMock(return_value=fake_es)

        with patch.object(elasticsearch, "_get_instance", fake_instance):
            self.assertIs(
                elasticsearch.index_put_settings(index="foo", body=body), None
            )

    def test_index_put_settings_failure(self):
        """
        Test if settings put failed with CommandExecutionError
        """

        body = {"settings": {"index": {"number_of_replicas": 2}}}
        fake_es = MagicMock()
        fake_es.indices = MagicMock()
        fake_es.indices.put_settings = MagicMock()
        fake_es.indices.put_settings.side_effect = TransportError("custom error", 123)
        fake_instance = MagicMock(return_value=fake_es)

        with patch.object(elasticsearch, "_get_instance", fake_instance):
            self.assertRaises(
                CommandExecutionError,
                elasticsearch.index_put_settings,
                index="foo",
                body=body,
            )

    # 'index_get' function tests: 3

    def test_index_get(self):
        """
        Test if index can be obtained
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def get(self, index=None):
                """
                Mock of index method
                """
                return {"test": "key"}

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertDictEqual(elasticsearch.index_get("foo", "bar"), {"test": "key"})

    def test_index_get_not(self):
        """
        Test if index doesn't exist
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def get(self, index=None):
                """
                Mock of index method
                """
                raise NotFoundError

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertIs(elasticsearch.index_get("foo", "bar"), None)

    def test_index_get_failure(self):
        """
        Test if index obtain fails with CommandExecutionError
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def get(self, index=None):
                """
                Mock of index method
                """
                raise TransportError("custom message", 123)

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertRaises(
                CommandExecutionError, elasticsearch.index_get, "foo", "bar"
            )

    # 'index_open' function tests: 3

    def test_index_open(self):
        """
        Test if index can be opened
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def open(
                self,
                index=None,
                allow_no_indices=None,
                expand_wildcards=None,
                ignore_unavailable=None,
            ):
                """
                Mock of index method
                """
                return {"acknowledged": True}

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertTrue(elasticsearch.index_open("foo", "bar"))

    def test_index_open_not(self):
        """
        Test if index open isn't acked
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def open(
                self,
                index=None,
                allow_no_indices=None,
                expand_wildcards=None,
                ignore_unavailable=None,
            ):
                """
                Mock of index method
                """
                return {"acknowledged": False}

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertFalse(elasticsearch.index_open("foo", "bar"))

    def test_index_open_failure(self):
        """
        Test if alias opening fails with CommandExecutionError
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def open(
                self,
                index=None,
                allow_no_indices=None,
                expand_wildcards=None,
                ignore_unavailable=None,
            ):
                """
                Mock of index method
                """
                raise TransportError("custom message", 123)

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertRaises(
                CommandExecutionError, elasticsearch.index_open, "foo", "bar"
            )

    # 'index_close' function tests: 3

    def test_index_close(self):
        """
        Test if index can be closed
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def close(
                self,
                index=None,
                allow_no_indices=None,
                expand_wildcards=None,
                ignore_unavailable=None,
            ):
                """
                Mock of index method
                """
                return {"acknowledged": True}

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertTrue(elasticsearch.index_close("foo", "bar"))

    def test_index_close_not(self):
        """
        Test if index close isn't acked
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def close(
                self,
                index=None,
                allow_no_indices=None,
                expand_wildcards=None,
                ignore_unavailable=None,
            ):
                """
                Mock of index method
                """
                return {"acknowledged": False}

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertFalse(elasticsearch.index_close("foo", "bar"))

    def test_index_close_failure(self):
        """
        Test if index closing fails with CommandExecutionError
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def close(
                self,
                index=None,
                allow_no_indices=None,
                expand_wildcards=None,
                ignore_unavailable=None,
            ):
                """
                Mock of index method
                """
                raise TransportError("custom message", 123)

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertRaises(
                CommandExecutionError, elasticsearch.index_close, "foo", "bar"
            )

    # 'mapping_create' function tests: 3

    def test_mapping_create(self):
        """
        Test if mapping can be created
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def put_mapping(self, index=None, doc_type=None, body=None):
                """
                Mock of put_mapping method
                """
                return {"acknowledged": True}

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertTrue(elasticsearch.mapping_create("foo", "bar", "baz"))

    def test_mapping_create_not(self):
        """
        Test if mapping creation didn't ack
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def put_mapping(self, index=None, doc_type=None, body=None):
                """
                Mock of put_mapping method
                """
                return {"acknowledged": False}

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertFalse(elasticsearch.mapping_create("foo", "bar", "baz"))

    def test_mapping_create_failure(self):
        """
        Test if mapping creation fails
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def put_mapping(self, index=None, doc_type=None, body=None):
                """
                Mock of put_mapping method
                """
                raise TransportError("custom message", 123)

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertRaises(
                CommandExecutionError, elasticsearch.mapping_create, "foo", "bar", "baz"
            )

    # 'mapping_delete' function tests: 3

    def test_mapping_delete(self):
        """
        Test if mapping can be created
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def delete_mapping(self, index=None, doc_type=None):
                """
                Mock of put_mapping method
                """
                return {"acknowledged": True}

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertTrue(elasticsearch.mapping_delete("foo", "bar", "baz"))

    def test_mapping_delete_not(self):
        """
        Test if mapping creation didn't ack
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def delete_mapping(self, index=None, doc_type=None):
                """
                Mock of put_mapping method
                """
                return {"acknowledged": False}

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertFalse(elasticsearch.mapping_delete("foo", "bar", "baz"))

    def test_mapping_delete_failure(self):
        """
        Test if mapping creation fails
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def delete_mapping(self, index=None, doc_type=None):
                """
                Mock of put_mapping method
                """
                raise TransportError("custom message", 123)

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertRaises(
                CommandExecutionError, elasticsearch.mapping_delete, "foo", "bar", "baz"
            )

    # 'mapping_get' function tests: 3

    def test_mapping_get(self):
        """
        Test if mapping can be created
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def get_mapping(self, index=None, doc_type=None):
                """
                Mock of get_mapping method
                """
                return {"test": "key"}

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertDictEqual(
                elasticsearch.mapping_get("foo", "bar", "baz"), {"test": "key"}
            )

    def test_mapping_get_not(self):
        """
        Test if mapping creation didn't ack
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def get_mapping(self, index=None, doc_type=None):
                """
                Mock of get_mapping method
                """
                raise NotFoundError

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertIs(elasticsearch.mapping_get("foo", "bar", "baz"), None)

    def test_mapping_get_failure(self):
        """
        Test if mapping creation fails
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def get_mapping(self, index=None, doc_type=None):
                """
                Mock of get_mapping method
                """
                raise TransportError("custom message", 123)

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertRaises(
                CommandExecutionError, elasticsearch.mapping_get, "foo", "bar", "baz"
            )

    # 'index_template_create' function tests: 3

    def test_index_template_create(self):
        """
        Test if mapping can be created
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def put_template(self, name=None, body=None):
                """
                Mock of put_template method
                """
                return {"acknowledged": True}

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertTrue(elasticsearch.index_template_create("foo", "bar"))

    def test_index_template_create_not(self):
        """
        Test if mapping creation didn't ack
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def put_template(self, name=None, body=None):
                """
                Mock of put_template method
                """
                return {"acknowledged": False}

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertFalse(elasticsearch.index_template_create("foo", "bar"))

    def test_index_template_create_failure(self):
        """
        Test if mapping creation fails
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def put_template(self, name=None, body=None):
                """
                Mock of put_template method
                """
                raise TransportError("custom message", 123)

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertRaises(
                CommandExecutionError, elasticsearch.index_template_create, "foo", "bar"
            )

    # 'index_template_delete' function tests: 3

    def test_index_template_delete(self):
        """
        Test if mapping can be created
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def delete_template(self, name=None):
                """
                Mock of delete_template method
                """
                return {"acknowledged": True}

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertTrue(elasticsearch.index_template_delete("foo"))

    def test_index_template_delete_not(self):
        """
        Test if mapping creation didn't ack
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def delete_template(self, name=None):
                """
                Mock of delete_template method
                """
                return {"acknowledged": False}

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertFalse(elasticsearch.index_template_delete("foo"))

    def test_index_template_delete_failure(self):
        """
        Test if mapping creation fails
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def delete_template(self, name=None):
                """
                Mock of delete_template method
                """
                raise TransportError("custom message", 123)

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertRaises(
                CommandExecutionError, elasticsearch.index_template_delete, "foo"
            )

    # 'index_template_exists' function tests: 3

    def test_index_template_exists(self):
        """
        Test if mapping can be created
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def exists_template(self, name=None):
                """
                Mock of exists_template method
                """
                return True

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertTrue(elasticsearch.index_template_exists("foo"))

    def test_index_template_exists_not(self):
        """
        Test if mapping creation didn't ack
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def exists_template(self, name=None):
                """
                Mock of exists_template method
                """
                return False

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertFalse(elasticsearch.index_template_exists("foo"))

    def test_index_template_exists_failure(self):
        """
        Test if mapping creation fails
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def exists_template(self, name=None):
                """
                Mock of exists_template method
                """
                raise TransportError("custom message", 123)

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertRaises(
                CommandExecutionError, elasticsearch.index_template_exists, "foo"
            )

    # 'index_template_get' function tests: 3

    def test_index_template_get(self):
        """
        Test if mapping can be created
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def get_template(self, name=None):
                """
                Mock of get_template method
                """
                return {"test": "key"}

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertDictEqual(
                elasticsearch.index_template_get("foo"), {"test": "key"}
            )

    def test_index_template_get_not(self):
        """
        Test if mapping creation didn't ack
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def get_template(self, name=None):
                """
                Mock of get_template method
                """
                raise NotFoundError

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertIs(elasticsearch.index_template_get("foo"), None)

    def test_index_template_get_failure(self):
        """
        Test if mapping creation fails
        """

        class MockElasticIndices:
            """
            Mock of Elasticsearch IndicesClient
            """

            def get_template(self, name=None):
                """
                Mock of get_template method
                """
                raise TransportError("custom message", 123)

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            indices = MockElasticIndices()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertRaises(
                CommandExecutionError, elasticsearch.index_template_get, "foo"
            )

    # 'pipeline_get' function tests: 4

    def test_pipeline_get(self):
        """
        Test if mapping can be created
        """

        class MockElasticIngest:
            """
            Mock of Elasticsearch IndicesClient
            """

            def get_pipeline(self, id=None):
                """
                Mock of get_pipeline method
                """
                return {"test": "key"}

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            ingest = MockElasticIngest()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertDictEqual(elasticsearch.pipeline_get("foo"), {"test": "key"})

    def test_pipeline_get_not(self):
        """
        Test if mapping creation didn't ack
        """

        class MockElasticIngest:
            """
            Mock of Elasticsearch IndicesClient
            """

            def get_pipeline(self, id=None):
                """
                Mock of get_pipeline method
                """
                raise NotFoundError

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            ingest = MockElasticIngest()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertIs(elasticsearch.pipeline_get("foo"), None)

    def test_pipeline_get_failure(self):
        """
        Test if mapping creation fails
        """

        class MockElasticIngest:
            """
            Mock of Elasticsearch IndicesClient
            """

            def get_pipeline(self, id=None):
                """
                Mock of get_pipeline method
                """
                raise TransportError("custom message", 123)

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            ingest = MockElasticIngest()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertRaises(CommandExecutionError, elasticsearch.pipeline_get, "foo")

    def test_pipeline_get_wrong_version(self):
        """
        Test if mapping creation fails with CEE on invalid elasticsearch-py version
        """

        class MockElasticIngest:
            pass

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            ingest = MockElasticIngest()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertRaises(CommandExecutionError, elasticsearch.pipeline_get, "foo")

    # 'pipeline_delete' function tests: 4

    def test_pipeline_delete(self):
        """
        Test if mapping can be created
        """

        class MockElasticIngest:
            """
            Mock of Elasticsearch IndicesClient
            """

            def delete_pipeline(self, id=None):
                """
                Mock of delete_pipeline method
                """
                return {"acknowledged": True}

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            ingest = MockElasticIngest()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertTrue(elasticsearch.pipeline_delete("foo"))

    def test_pipeline_delete_not(self):
        """
        Test if mapping creation didn't ack
        """

        class MockElasticIngest:
            """
            Mock of Elasticsearch IndicesClient
            """

            def delete_pipeline(self, id=None):
                """
                Mock of delete_pipeline method
                """
                return {"acknowledged": False}

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            ingest = MockElasticIngest()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertFalse(elasticsearch.pipeline_delete("foo"))

    def test_pipeline_delete_failure(self):
        """
        Test if mapping creation fails
        """

        class MockElasticIngest:
            """
            Mock of Elasticsearch IndicesClient
            """

            def delete_pipeline(self, id=None):
                """
                Mock of delete_pipeline method
                """
                raise TransportError("custom message", 123)

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            ingest = MockElasticIngest()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertRaises(
                CommandExecutionError, elasticsearch.pipeline_delete, "foo"
            )

    def test_pipeline_delete_wrong_version(self):
        """
        Test if mapping creation fails with CEE on invalid elasticsearch-py version
        """

        class MockElasticIngest:
            pass

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            ingest = MockElasticIngest()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertRaises(
                CommandExecutionError, elasticsearch.pipeline_delete, "foo"
            )

    # 'pipeline_create' function tests: 4

    def test_pipeline_create(self):
        """
        Test if mapping can be created
        """

        class MockElasticIngest:
            """
            Mock of Elasticsearch IndicesClient
            """

            def put_pipeline(self, id=None, body=None):
                """
                Mock of put_pipeline method
                """
                return {"acknowledged": True}

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            ingest = MockElasticIngest()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertTrue(elasticsearch.pipeline_create("foo", "bar"))

    def test_pipeline_create_not(self):
        """
        Test if mapping creation didn't ack
        """

        class MockElasticIngest:
            """
            Mock of Elasticsearch IndicesClient
            """

            def put_pipeline(self, id=None, body=None):
                """
                Mock of put_pipeline method
                """
                return {"acknowledged": False}

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            ingest = MockElasticIngest()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertFalse(elasticsearch.pipeline_create("foo", "bar"))

    def test_pipeline_create_failure(self):
        """
        Test if mapping creation fails
        """

        class MockElasticIngest:
            """
            Mock of Elasticsearch IndicesClient
            """

            def put_pipeline(self, id=None, body=None):
                """
                Mock of put_pipeline method
                """
                raise TransportError("custom message", 123)

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            ingest = MockElasticIngest()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertRaises(
                CommandExecutionError, elasticsearch.pipeline_create, "foo", "bar"
            )

    def test_pipeline_create_wrong_version(self):
        """
        Test if mapping creation fails with CEE on invalid elasticsearch-py version
        """

        class MockElasticIngest:
            pass

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            ingest = MockElasticIngest()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertRaises(
                CommandExecutionError, elasticsearch.pipeline_create, "foo", "bar"
            )

    # 'pipeline_simulate' function tests: 3

    def test_pipeline_simulate(self):
        """
        Test if mapping can be created
        """

        class MockElasticIngest:
            """
            Mock of Elasticsearch IndicesClient
            """

            def simulate(self, id=None, body=None, verbose=None):
                """
                Mock of simulate method
                """
                return {"test": "key"}

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            ingest = MockElasticIngest()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertDictEqual(
                elasticsearch.pipeline_simulate("foo", "bar"), {"test": "key"}
            )

    def test_pipeline_simulate_failure(self):
        """
        Test if mapping creation fails
        """

        class MockElasticIngest:
            """
            Mock of Elasticsearch IndicesClient
            """

            def simulate(self, id=None, body=None, verbose=None):
                """
                Mock of simulate method
                """
                raise TransportError("custom message", 123)

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            ingest = MockElasticIngest()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertRaises(
                CommandExecutionError, elasticsearch.pipeline_simulate, "foo", "bar"
            )

    def test_pipeline_simulate_wrong_version(self):
        """
        Test if mapping creation fails with CEE on invalid elasticsearch-py version
        """

        class MockElasticIngest:
            pass

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            ingest = MockElasticIngest()

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertRaises(
                CommandExecutionError, elasticsearch.pipeline_simulate, "foo", "bar"
            )

    # 'search_template_get' function tests: 3

    def test_search_template_get(self):
        """
        Test if mapping can be created
        """

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            def get_template(self, id=None):
                """
                Mock of get_template method
                """
                return {"test": "key"}

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertDictEqual(
                elasticsearch.search_template_get("foo"), {"test": "key"}
            )

    def test_search_template_get_not(self):
        """
        Test if mapping can be created
        """

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            def get_template(self, id=None):
                """
                Mock of get_template method
                """
                raise NotFoundError

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertIs(elasticsearch.search_template_get("foo"), None)

    def test_search_template_get_failure(self):
        """
        Test if mapping creation fails
        """

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            def get_template(self, id=None):
                """
                Mock of get_template method
                """
                raise TransportError("custom message", 123)

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertRaises(
                CommandExecutionError, elasticsearch.search_template_get, "foo"
            )

    # 'search_template_create' function tests: 3

    def test_search_template_create(self):
        """
        Test if mapping can be created
        """

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            def put_template(self, id=None, body=None):
                """
                Mock of put_template method
                """
                return {"acknowledged": True}

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertTrue(elasticsearch.search_template_create("foo", "bar"))

    def test_search_template_create_not(self):
        """
        Test if mapping can be created
        """

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            def put_template(self, id=None, body=None):
                """
                Mock of put_template method
                """
                return {"acknowledged": False}

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertFalse(elasticsearch.search_template_create("foo", "bar"))

    def test_search_template_create_failure(self):
        """
        Test if mapping creation fails
        """

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            def put_template(self, id=None, body=None):
                """
                Mock of put_template method
                """
                raise TransportError("custom message", 123)

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertRaises(
                CommandExecutionError,
                elasticsearch.search_template_create,
                "foo",
                "bar",
            )

    # 'search_template_delete' function tests: 4

    def test_search_template_delete(self):
        """
        Test if mapping can be deleted
        """

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            def delete_template(self, id=None):
                """
                Mock of delete_template method
                """
                return {"acknowledged": True}

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertTrue(elasticsearch.search_template_delete("foo"))

    def test_search_template_delete_not(self):
        """
        Test if mapping can be deleted but not acked
        """

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            def delete_template(self, id=None):
                """
                Mock of delete_template method
                """
                return {"acknowledged": False}

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertFalse(elasticsearch.search_template_delete("foo"))

    def test_search_template_delete_not_exists(self):
        """
        Test if deleting mapping doesn't exist
        """

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            def delete_template(self, id=None):
                """
                Mock of delete_template method
                """
                raise NotFoundError

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertTrue(elasticsearch.search_template_delete("foo"))

    def test_search_template_delete_failure(self):
        """
        Test if mapping deletion fails
        """

        class MockElastic:
            """
            Mock of Elasticsearch client
            """

            def delete_template(self, id=None):
                """
                Mock of delete_template method
                """
                raise TransportError("custom message", 123)

        with patch.object(
            elasticsearch, "_get_instance", MagicMock(return_value=MockElastic())
        ):
            self.assertRaises(
                CommandExecutionError, elasticsearch.search_template_delete, "foo"
            )

    # Cluster settings tests below.
    # We're assuming that _get_instance is properly tested
    # These tests are very simple in nature, mostly checking default arguments.
    def test_cluster_get_settings_succeess(self):
        """
        Test if cluster get_settings fetch succeeds
        """

        expected_settings = {"transient": {}, "persistent": {}}
        fake_es = MagicMock()
        fake_es.cluster = MagicMock()
        fake_es.cluster.get_settings = MagicMock(return_value=expected_settings)
        fake_instance = MagicMock(return_value=fake_es)

        with patch.object(elasticsearch, "_get_instance", fake_instance):
            actual_settings = elasticsearch.cluster_get_settings()
            fake_es.cluster.get_settings.assert_called_with(
                flat_settings=False, include_defaults=False
            )
            assert actual_settings == expected_settings

    def test_cluster_get_settings_failure(self):
        """
        Test if cluster get_settings fetch fails with CommandExecutionError
        """

        fake_es = MagicMock()
        fake_es.cluster = MagicMock()
        fake_es.cluster.get_settings = MagicMock()
        fake_es.cluster.get_settings.side_effect = TransportError("custom error", 123)
        fake_instance = MagicMock(return_value=fake_es)

        with patch.object(elasticsearch, "_get_instance", fake_instance):
            self.assertRaises(CommandExecutionError, elasticsearch.cluster_get_settings)

    def test_cluster_put_settings_succeess(self):
        """
        Test if cluster put_settings succeeds
        """

        expected_settings = {
            "acknowledged": True,
            "transient": {},
            "persistent": {"indices": {"recovery": {"max_bytes_per_sec": "50mb"}}},
        }
        body = {
            "transient": {},
            "persistent": {"indices.recovery.max_bytes_per_sec": "50mb"},
        }
        fake_es = MagicMock()
        fake_es.cluster = MagicMock()
        fake_es.cluster.put_settings = MagicMock(return_value=expected_settings)
        fake_instance = MagicMock(return_value=fake_es)

        with patch.object(elasticsearch, "_get_instance", fake_instance):
            actual_settings = elasticsearch.cluster_put_settings(body=body)
            fake_es.cluster.put_settings.assert_called_with(
                body=body, flat_settings=False
            )
            assert actual_settings == expected_settings

    def test_cluster_put_settings_failure(self):
        """
        Test if cluster put_settings fails with CommandExecutionError
        """

        body = {
            "transient": {},
            "persistent": {"indices.recovery.max_bytes_per_sec": "50mb"},
        }
        fake_es = MagicMock()
        fake_es.cluster = MagicMock()
        fake_es.cluster.put_settings = MagicMock()
        fake_es.cluster.put_settings.side_effect = TransportError("custom error", 123)
        fake_instance = MagicMock(return_value=fake_es)

        with patch.object(elasticsearch, "_get_instance", fake_instance):
            self.assertRaises(
                CommandExecutionError, elasticsearch.cluster_put_settings, body=body
            )

    def test_cluster_put_settings_nobody(self):
        """
        Test if cluster put_settings fails with SaltInvocationError
        """

        self.assertRaises(SaltInvocationError, elasticsearch.cluster_put_settings)

    # flush_synced tests below.
    # We're assuming that _get_instance is properly tested
    # These tests are very simple in nature, mostly checking default arguments.
    def test_flush_synced_succeess(self):
        """
        Test if flush_synced succeeds
        """

        expected_return = {"_shards": {"failed": 0, "successful": 0, "total": 0}}
        fake_es = MagicMock()
        fake_es.indices = MagicMock()
        fake_es.indices.flush_synced = MagicMock(return_value=expected_return)
        fake_instance = MagicMock(return_value=fake_es)

        with patch.object(elasticsearch, "_get_instance", fake_instance):
            output = elasticsearch.flush_synced(
                index="_all",
                ignore_unavailable=True,
                allow_no_indices=True,
                expand_wildcards="all",
            )
            fake_es.indices.flush_synced.assert_called_with(
                {
                    "index": "_all",
                    "ignore_unavailable": True,
                    "allow_no_indices": True,
                    "expand_wildcards": "all",
                }
            )
            assert output == expected_return

    def test_flush_synced_failure(self):
        """
        Test if flush_synced fails with CommandExecutionError
        """

        fake_es = MagicMock()
        fake_es.indices = MagicMock()
        fake_es.indices.flush_synced = MagicMock()
        fake_es.indices.flush_synced.side_effect = TransportError("custom error", 123)
        fake_instance = MagicMock(return_value=fake_es)

        with patch.object(elasticsearch, "_get_instance", fake_instance):
            self.assertRaises(CommandExecutionError, elasticsearch.flush_synced)



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest
import salt.modules.guestfs as guestfs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, call, patch
from tests.support.unit import TestCase


class GuestfsTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.guestfs
    """

    def setup_loader_modules(self):
        return {guestfs: {}}

    # 'mount' function tests: 1
    def test_mount(self):
        """
        Test if it mounts an image
        """
        # Test case with non-existing mount folder
        run_mock = MagicMock(return_value="")
        with patch(
            "os.path.join", MagicMock(return_value="/tmp/guest/fedora.qcow")
        ), patch("os.path.isdir", MagicMock(return_value=False)), patch(
            "os.makedirs", MagicMock()
        ) as makedirs_mock, patch(
            "os.listdir", MagicMock(return_value=False)
        ), patch.dict(
            guestfs.__salt__, {"cmd.run": run_mock}
        ):
            self.assertTrue(guestfs.mount("/srv/images/fedora.qcow"))
            run_mock.assert_called_once_with(
                "guestmount -i -a /srv/images/fedora.qcow --rw /tmp/guest/fedora.qcow",
                python_shell=False,
            )
            makedirs_mock.assert_called_once()

        # Test case with existing but empty mount folder
        run_mock.reset_mock()
        with patch(
            "os.path.join", MagicMock(return_value="/tmp/guest/fedora.qcow")
        ), patch("os.path.isdir", MagicMock(return_value=True)), patch(
            "os.makedirs", MagicMock()
        ) as makedirs_mock, patch(
            "os.listdir", MagicMock(return_value=False)
        ), patch.dict(
            guestfs.__salt__, {"cmd.run": run_mock}
        ):
            self.assertTrue(guestfs.mount("/srv/images/fedora.qcow"))
            run_mock.assert_called_once_with(
                "guestmount -i -a /srv/images/fedora.qcow --rw /tmp/guest/fedora.qcow",
                python_shell=False,
            )
            makedirs_mock.assert_not_called()

        # Test case with existing but not empty mount folder
        run_mock.reset_mock()
        with patch(
            "os.path.join",
            MagicMock(
                side_effect=["/tmp/guest/fedora.qcow", "/tmp/guest/fedora.qcowabc"]
            ),
        ), patch("os.path.isdir", MagicMock(side_effect=[True, False])), patch(
            "os.makedirs", MagicMock()
        ) as makedirs_mock, patch(
            "os.listdir", MagicMock(side_effect=[True, False])
        ), patch.dict(
            guestfs.__salt__, {"cmd.run": run_mock}
        ):
            self.assertTrue(guestfs.mount("/srv/images/fedora.qcow"))
            run_mock.assert_called_once_with(
                "guestmount -i -a /srv/images/fedora.qcow --rw"
                " /tmp/guest/fedora.qcowabc",
                python_shell=False,
            )
            makedirs_mock.assert_called_once()

    @pytest.mark.slow_test
    def test_umount(self):
        """
        Test the guestfs.unmount function
        """
        run_mock = MagicMock(side_effect=["", "lsof output line", ""])
        with patch.dict(guestfs.__salt__, {"cmd.run": run_mock}):
            guestfs.umount("/tmp/mnt/opensuse.qcow", disk="/path/to/opensuse.qcow")
            expected = [
                call("guestunmount -q /tmp/mnt/opensuse.qcow"),
                call("lsof /path/to/opensuse.qcow"),
                call("lsof /path/to/opensuse.qcow"),
            ]
            self.assertEqual(expected, run_mock.call_args_list)

        # Test without the disk image path
        run_mock = MagicMock(side_effect=[""])
        with patch.dict(guestfs.__salt__, {"cmd.run": run_mock}):
            guestfs.umount("/tmp/mnt/opensuse.qcow")
            expected = [call("guestunmount -q /tmp/mnt/opensuse.qcow")]
            self.assertEqual(expected, run_mock.call_args_list)



import os
import textwrap

import pytest
import salt.utils.files
import salt.utils.platform
from tests.support.case import ModuleCase
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import skipIf


@pytest.mark.windows_whitelisted
class EnabledTest(ModuleCase):
    """
    validate the use of shell processing for cmd.run on the salt command line
    and in templating
    """

    cmd = (
        "printf '%s\\n' first second third | wc -l ; "
        "export SALTY_VARIABLE='saltines' && echo $SALTY_VARIABLE ; "
        "echo duh &> /dev/null"
    )

    @skipIf(salt.utils.platform.is_windows(), "Skip on Windows OS")
    @skipIf(salt.utils.platform.is_freebsd(), "Skip on FreeBSD")
    def test_shell_default_enabled(self):
        """
        ensure that python_shell defaults to True for cmd.run
        """
        enabled_ret = "3\nsaltines"  # the result of running self.cmd in a shell
        ret = self.run_function("cmd.run", [self.cmd])
        self.assertEqual(ret.strip(), enabled_ret)

    @skipIf(salt.utils.platform.is_windows(), "Skip on Windows OS")
    def test_shell_disabled(self):
        """
        test shell disabled output for cmd.run
        """
        disabled_ret = (
            "first\nsecond\nthird\n|\nwc\n-l\n;\nexport\nSALTY_VARIABLE=saltines"
            "\n&&\necho\n$SALTY_VARIABLE\n;\necho\nduh\n&>\n/dev/null"
        )
        ret = self.run_function("cmd.run", [self.cmd], python_shell=False)
        self.assertEqual(ret, disabled_ret)

    @skipIf(salt.utils.platform.is_windows(), "Skip on Windows OS")
    @skipIf(salt.utils.platform.is_freebsd(), "Skip on FreeBSD")
    def test_template_shell(self):
        """
        Test cmd.shell works correctly when using a template.

        Note: This test used to test that python_shell defaulted to True for templates
        in releases before 2017.7.0. The cmd.run --> cmd.shell aliasing was removed in
        2017.7.0. Templates should now be using cmd.shell.
        """
        state_name = "template_shell_enabled"
        state_filename = state_name + ".sls"
        state_file = os.path.join(RUNTIME_VARS.BASE_FILES, state_filename)

        enabled_ret = "3 saltines"  # the result of running self.cmd in a shell
        ret_key = "test_|-shell_enabled_|-{}_|-configurable_test_state".format(
            enabled_ret
        )

        try:
            with salt.utils.files.fopen(state_file, "w") as fp_:
                fp_.write(
                    textwrap.dedent(
                        """\
                {{% set shell_enabled = salt['cmd.shell']("{0}").strip() %}}

                shell_enabled:
                  test.configurable_test_state:
                    - name: '{{{{ shell_enabled }}}}'
                """.format(
                            self.cmd
                        )
                    )
                )

            ret = self.run_function("state.sls", [state_name])
            self.assertEqual(ret[ret_key]["name"], enabled_ret)
        finally:
            os.remove(state_file)

    @skipIf(salt.utils.platform.is_windows(), "Skip on Windows OS")
    @pytest.mark.slow_test
    def test_template_default_disabled(self):
        """
        test shell disabled output for templates (python_shell=False is the default
        beginning with the 2017.7.0 release).
        """
        state_name = "template_shell_disabled"
        state_filename = state_name + ".sls"
        state_file = os.path.join(RUNTIME_VARS.BASE_FILES, state_filename)

        # the result of running self.cmd not in a shell
        disabled_ret = (
            "first second third | wc -l ; export SALTY_VARIABLE=saltines "
            "&& echo $SALTY_VARIABLE ; echo duh &> /dev/null"
        )
        ret_key = "test_|-shell_enabled_|-{}_|-configurable_test_state".format(
            disabled_ret
        )

        try:
            with salt.utils.files.fopen(state_file, "w") as fp_:
                fp_.write(
                    textwrap.dedent(
                        """\
                {{% set shell_disabled = salt['cmd.run']("{0}") %}}

                shell_enabled:
                  test.configurable_test_state:
                    - name: '{{{{ shell_disabled }}}}'
                """.format(
                            self.cmd
                        )
                    )
                )

            ret = self.run_function("state.sls", [state_name])
            self.assertEqual(ret[ret_key]["name"], disabled_ret)
        finally:
            os.remove(state_file)



"""
Device-Mapper module
"""

import os.path


def multipath_list():
    """
    Device-Mapper Multipath list

    CLI Example:

    .. code-block:: bash

        salt '*' devmap.multipath_list
    """
    cmd = "multipath -l"
    return __salt__["cmd.run"](cmd).splitlines()


def multipath_flush(device):
    """
    Device-Mapper Multipath flush

    CLI Example:

    .. code-block:: bash

        salt '*' devmap.multipath_flush mpath1
    """
    if not os.path.exists(device):
        return "{} does not exist".format(device)

    cmd = "multipath -f {}".format(device)
    return __salt__["cmd.run"](cmd).splitlines()



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""
import salt.modules.sdb as sdb
from salt.exceptions import SaltInvocationError
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.unit import TestCase


class SdbTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.sdb
    """

    def setup_loader_modules(self):
        return {sdb: {}}

    # 'get' function tests: 4

    def test_get(self):
        """
        Test if it gets a value from a db, using a uri in the form of
        sdb://<profile>/<key>
        """
        self.assertEqual(sdb.get("sdb://salt/foo"), "sdb://salt/foo")

    def test_get_strict_no_sdb_in_uri(self):
        """
        Test if SaltInvocationError exception will be raised if we
        don't start uri with sdb://
        """

        msg = 'SDB uri must start with "sdb://"'
        with self.assertRaisesRegex(SaltInvocationError, msg) as cm:
            sdb.get("://salt/foo", strict=True)

    def test_get_strict_no_profile(self):
        """
        Test if SaltInvocationError exception will be raised if we
        don't have a valid profile in the uri
        """

        msg = "SDB uri must have a profile name as a first part of the uri before the /"
        with self.assertRaisesRegex(SaltInvocationError, msg) as cm:
            sdb.get("sdb://salt", strict=True)

    def test_get_strict_no_profile_in_config(self):
        """
        Test if SaltInvocationError exception will be raised if we
        don't have expected profile in the minion config
        """

        msg = 'SDB profile "salt" wasnt found in the minion configuration'
        with self.assertRaisesRegex(SaltInvocationError, msg) as cm:
            sdb.get("sdb://salt/foo", strict=True)

    # 'set_' function tests: 1

    def test_set(self):
        """
        Test if it sets a value from a db, using a uri in the form of
        sdb://<profile>/<key>
        """
        self.assertFalse(sdb.set_("sdb://mymemcached/foo", "bar"))



"""
tests for pkg state
"""

import logging
import os
import time

import pytest
import salt.utils.files
import salt.utils.path
import salt.utils.pkg.rpm
import salt.utils.platform
from tests.support.case import ModuleCase
from tests.support.helpers import not_runs_on, requires_system_grains, runs_on
from tests.support.mixins import SaltReturnAssertsMixin
from tests.support.unit import skipIf

log = logging.getLogger(__name__)


@pytest.mark.windows_whitelisted
@pytest.mark.destructive_test
class PkgTest(ModuleCase, SaltReturnAssertsMixin):
    _PKG_EPOCH_TARGETS = []
    _PKG_32_TARGETS = []
    _PKG_CAP_TARGETS = []
    _PKG_DOT_TARGETS = []
    _WILDCARDS_SUPPORTED = False
    _VERSION_SPEC_SUPPORTED = True

    @classmethod
    @requires_system_grains
    def setUpClass(cls, grains=None):  # pylint:disable=W0221
        cls.ctx = {}
        cls._PKG_TARGETS = ["figlet", "sl"]
        if grains["os"] == "Windows":
            cls._PKG_TARGETS = ["vlc", "putty"]
        elif grains["os"] == "FreeBSD":
            cls._VERSION_SPEC_SUPPORTED = False
        elif grains["os_family"] in ("Arch", "Debian"):
            cls._WILDCARDS_SUPPORTED = True
        elif grains["os"] == "Amazon":
            cls._PKG_TARGETS = ["lynx", "gnuplot"]
        elif grains["os_family"] == "RedHat":
            cls._PKG_TARGETS = ["units", "zsh-html"]
            cls._WILDCARDS_SUPPORTED = True
            if grains["os"] == "CentOS":
                if grains["osmajorrelease"] == 5:
                    cls._PKG_32_TARGETS = ["xz-devel.i386"]
                else:
                    cls._PKG_32_TARGETS.append("xz-devel.i686")
            if grains["osmajorrelease"] == 5:
                cls._PKG_DOT_TARGETS = ["python-migrate0.5"]
            elif grains["osmajorrelease"] == 6:
                cls._PKG_DOT_TARGETS = ["tomcat6-el-2.1-api"]
            elif grains["osmajorrelease"] == 7:
                cls._PKG_DOT_TARGETS = ["tomcat-el-2.2-api"]
                cls._PKG_EPOCH_TARGETS = ["comps-extras"]
            elif grains["osmajorrelease"] == 8:
                cls._PKG_DOT_TARGETS = ["vid.stab"]
                cls._PKG_EPOCH_TARGETS = ["traceroute"]
        elif grains["os_family"] == "Suse":
            cls._PKG_TARGETS = ["lynx", "htop"]
            if grains["os"] == "SUSE":
                cls._PKG_CAP_TARGETS = [("perl(ZNC)", "znc-perl")]

    @classmethod
    def tearDownClass(cls):
        del cls.ctx

    def latest_version(self, *names):
        """
        Helper function which ensures that we don't make any unnecessary calls to
        pkg.latest_version to figure out what version we need to install. This
        won't stop pkg.latest_version from being run in a pkg.latest state, but it
        will reduce the amount of times we check the latest version here in the
        test suite.
        """
        key = "latest_version"
        if key not in self.ctx:
            self.ctx[key] = dict()
        targets = [x for x in names if x not in self.ctx[key]]
        if targets:
            result = self.run_function("pkg.latest_version", targets, refresh=False)
            try:
                self.ctx[key].update(result)
            except ValueError:
                # Only a single target, pkg.latest_version returned a string
                self.ctx[key][targets[0]] = result

        ret = {x: self.ctx[key].get(x, "") for x in names}
        if len(names) == 1:
            return ret[names[0]]
        return ret

    @requires_system_grains
    def setUp(self, grains=None):  # pylint:disable=W0221
        super().setUp()
        if "refresh" not in self.ctx:
            self.run_function("pkg.refresh_db")
            self.ctx["refresh"] = True

        # If this is Arch Linux, check if pacman is in use by another process
        if grains["os_family"] == "Arch":
            for _ in range(12):
                if not os.path.isfile("/var/lib/pacman/db.lck"):
                    break
                else:
                    time.sleep(5)
            else:
                raise Exception("Package database locked after 60 seconds, bailing out")

    @pytest.mark.requires_salt_modules("pkg.version")
    @pytest.mark.requires_salt_states("pkg.installed", "pkg.removed")
    @pytest.mark.slow_test
    def test_pkg_001_installed(self):
        """
        This is a destructive test as it installs and then removes a package
        """
        target = self._PKG_TARGETS[0]
        version = self.run_function("pkg.version", [target])

        # If this assert fails, we need to find new targets, this test needs to
        # be able to test successful installation of packages, so this package
        # needs to not be installed before we run the states below
        self.assertFalse(version)

        ret = self.run_state("pkg.installed", name=target, refresh=False)
        self.assertSaltTrueReturn(ret)
        ret = self.run_state("pkg.removed", name=target)
        self.assertSaltTrueReturn(ret)

    @pytest.mark.requires_salt_states("pkg.installed", "pkg.removed")
    def test_pkg_002_installed_with_version(self):
        """
        This is a destructive test as it installs and then removes a package
        """
        if not self._VERSION_SPEC_SUPPORTED:
            self.skipTest("Version specification not supported")
        target = self._PKG_TARGETS[0]
        version = self.latest_version(target)

        # If this assert fails, we need to find new targets, this test needs to
        # be able to test successful installation of packages, so this package
        # needs to not be installed before we run the states below
        self.assertTrue(version)

        ret = self.run_state(
            "pkg.installed", name=target, version=version, refresh=False
        )
        self.assertSaltTrueReturn(ret)
        ret = self.run_state("pkg.removed", name=target)
        self.assertSaltTrueReturn(ret)

    @pytest.mark.requires_salt_states("pkg.installed", "pkg.removed")
    @pytest.mark.slow_test
    def test_pkg_003_installed_multipkg(self):
        """
        This is a destructive test as it installs and then removes two packages
        """
        version = self.run_function("pkg.version", self._PKG_TARGETS)

        # If this assert fails, we need to find new targets, this test needs to
        # be able to test successful installation of packages, so these
        # packages need to not be installed before we run the states below
        self.assertFalse(any(version.values()))
        self.assertSaltTrueReturn(
            self.run_state("pkg.removed", name=None, pkgs=self._PKG_TARGETS)
        )

        try:
            ret = self.run_state(
                "pkg.installed", name=None, pkgs=self._PKG_TARGETS, refresh=False
            )
            self.assertSaltTrueReturn(ret)
        finally:
            ret = self.run_state("pkg.removed", name=None, pkgs=self._PKG_TARGETS)
            self.assertSaltTrueReturn(ret)

    @pytest.mark.requires_salt_states("pkg.installed", "pkg.removed")
    def test_pkg_004_installed_multipkg_with_version(self):
        """
        This is a destructive test as it installs and then removes two packages
        """
        if not self._VERSION_SPEC_SUPPORTED:
            self.skipTest("Version specification not supported")
        version = self.latest_version(self._PKG_TARGETS[0])

        # If this assert fails, we need to find new targets, this test needs to
        # be able to test successful installation of packages, so these
        # packages need to not be installed before we run the states below
        self.assertTrue(bool(version))

        pkgs = [{self._PKG_TARGETS[0]: version}, self._PKG_TARGETS[1]]

        try:
            ret = self.run_state("pkg.installed", name=None, pkgs=pkgs, refresh=False)
            self.assertSaltTrueReturn(ret)
        finally:
            ret = self.run_state("pkg.removed", name=None, pkgs=self._PKG_TARGETS)
            self.assertSaltTrueReturn(ret)

    @pytest.mark.requires_salt_modules("pkg.version")
    @pytest.mark.requires_salt_states("pkg.installed", "pkg.removed")
    def test_pkg_005_installed_32bit(self):
        """
        This is a destructive test as it installs and then removes a package
        """
        if not self._PKG_32_TARGETS:
            self.skipTest("No 32 bit packages have been specified for testing")

        target = self._PKG_32_TARGETS[0]

        # _PKG_TARGETS_32 is only populated for platforms for which Salt has to
        # munge package names for 32-bit-on-x86_64 (Currently only Ubuntu and
        # RHEL-based). Don't actually perform this test on other platforms.
        version = self.run_function("pkg.version", [target])

        # If this assert fails, we need to find a new target. This test
        # needs to be able to test successful installation of packages, so
        # the target needs to not be installed before we run the states
        # below
        self.assertFalse(version)

        ret = self.run_state("pkg.installed", name=target, refresh=False)
        self.assertSaltTrueReturn(ret)
        ret = self.run_state("pkg.removed", name=target)
        self.assertSaltTrueReturn(ret)

    @pytest.mark.requires_salt_states("pkg.installed", "pkg.removed")
    def test_pkg_006_installed_32bit_with_version(self):
        """
        This is a destructive test as it installs and then removes a package
        """
        if not self._PKG_32_TARGETS:
            self.skipTest("No 32 bit packages have been specified for testing")

        target = self._PKG_32_TARGETS[0]

        # _PKG_TARGETS_32 is only populated for platforms for which Salt has to
        # munge package names for 32-bit-on-x86_64 (Currently only Ubuntu and
        # RHEL-based). Don't actually perform this test on other platforms.
        version = self.latest_version(target)

        # If this assert fails, we need to find a new target. This test
        # needs to be able to test successful installation of the package, so
        # the target needs to not be installed before we run the states
        # below
        self.assertTrue(version)

        ret = self.run_state(
            "pkg.installed", name=target, version=version, refresh=False
        )
        self.assertSaltTrueReturn(ret)
        ret = self.run_state("pkg.removed", name=target)
        self.assertSaltTrueReturn(ret)

    @pytest.mark.requires_salt_states("pkg.installed", "pkg.removed")
    def test_pkg_007_with_dot_in_pkgname(self=None):
        """
        This tests for the regression found in the following issue:
        https://github.com/saltstack/salt/issues/8614

        This is a destructive test as it installs a package
        """
        if not self._PKG_DOT_TARGETS:
            self.skipTest(
                'No packages with "." in their name have been specified',
            )

        target = self._PKG_DOT_TARGETS[0]

        version = self.latest_version(target)
        # If this assert fails, we need to find a new target. This test
        # needs to be able to test successful installation of the package, so
        # the target needs to not be installed before we run the
        # pkg.installed state below
        self.assertTrue(bool(version))
        ret = self.run_state("pkg.installed", name=target, refresh=False)
        self.assertSaltTrueReturn(ret)
        ret = self.run_state("pkg.removed", name=target)
        self.assertSaltTrueReturn(ret)

    @pytest.mark.requires_salt_states("pkg.installed", "pkg.removed")
    def test_pkg_008_epoch_in_version(self):
        """
        This tests for the regression found in the following issue:
        https://github.com/saltstack/salt/issues/8614

        This is a destructive test as it installs a package
        """
        if not self._PKG_EPOCH_TARGETS:
            self.skipTest('No targets have been configured with "epoch" in the version')

        target = self._PKG_EPOCH_TARGETS[0]

        version = self.latest_version(target)
        # If this assert fails, we need to find a new target. This test
        # needs to be able to test successful installation of the package, so
        # the target needs to not be installed before we run the
        # pkg.installed state below
        self.assertTrue(version)
        ret = self.run_state(
            "pkg.installed", name=target, version=version, refresh=False
        )
        self.assertSaltTrueReturn(ret)
        ret = self.run_state("pkg.removed", name=target)
        self.assertSaltTrueReturn(ret)

    @pytest.mark.requires_salt_modules("pkg.version", "pkg.info_installed")
    @pytest.mark.requires_salt_states("pkg.installed", "pkg.removed")
    @runs_on(kernel="linux")
    @not_runs_on(os="Amazon")
    @pytest.mark.slow_test
    def test_pkg_009_latest_with_epoch(self):
        """
        This tests for the following issue:
        https://github.com/saltstack/salt/issues/31014

        This is a destructive test as it installs a package
        """
        package = "bash-completion"
        pkgquery = "version"

        ret = self.run_state("pkg.installed", name=package, refresh=False)
        self.assertSaltTrueReturn(ret)

        ret = self.run_function("pkg.info_installed", [package])
        self.assertTrue(pkgquery in str(ret))

    @pytest.mark.requires_salt_states("pkg.latest", "pkg.removed")
    @pytest.mark.slow_test
    def test_pkg_010_latest(self):
        """
        This tests pkg.latest with a package that has no epoch (or a zero
        epoch).
        """
        target = self._PKG_TARGETS[0]
        version = self.latest_version(target)

        # If this assert fails, we need to find new targets, this test needs to
        # be able to test successful installation of packages, so this package
        # needs to not be installed before we run the states below
        self.assertTrue(version)

        ret = self.run_state("pkg.latest", name=target, refresh=False)
        self.assertSaltTrueReturn(ret)
        ret = self.run_state("pkg.removed", name=target)
        self.assertSaltTrueReturn(ret)

    @pytest.mark.requires_salt_modules(
        "pkg.list_pkgs", "pkg.list_upgrades", "pkg.version"
    )
    @pytest.mark.requires_salt_states("pkg.latest")
    @runs_on(kernel="linux", os_family="Debian")
    @pytest.mark.slow_test
    def test_pkg_011_latest_only_upgrade(self):
        """
        WARNING: This test will pick a package with an available upgrade (if
        there is one) and upgrade it to the latest version.
        """
        target = self._PKG_TARGETS[0]

        # If this assert fails, we need to find new targets, this test needs to
        # be able to test that the state fails when you try to run the state
        # with only_upgrade=True on a package which is not already installed,
        # so the targeted package needs to not be installed before we run the
        # state below.
        version = self.latest_version(target)
        self.assertTrue(version)

        ret = self.run_state(
            "pkg.latest", name=target, refresh=False, only_upgrade=True
        )
        self.assertSaltFalseReturn(ret)

        # Now look for updates and try to run the state on a package which is already up-to-date.
        installed_pkgs = self.run_function("pkg.list_pkgs")
        updates = self.run_function("pkg.list_upgrades", refresh=False)

        for pkgname in updates:
            if pkgname in installed_pkgs:
                target = pkgname
                break
        else:
            target = ""
            log.warning(
                "No available upgrades to installed packages, skipping "
                "only_upgrade=True test with already-installed package. For "
                "best results run this test on a machine with upgrades "
                "available."
            )

        if target:
            ret = self.run_state(
                "pkg.latest", name=target, refresh=False, only_upgrade=True
            )
            self.assertSaltTrueReturn(ret)
            new_version = self.run_function("pkg.version", [target])
            self.assertEqual(new_version, updates[target])
            ret = self.run_state(
                "pkg.latest", name=target, refresh=False, only_upgrade=True
            )
            self.assertEqual(
                ret["pkg_|-{0}_|-{0}_|-latest".format(target)]["comment"],
                "Package {} is already up-to-date".format(target),
            )

    @pytest.mark.requires_salt_modules("pkg.version")
    @pytest.mark.requires_salt_states("pkg.installed", "pkg.removed")
    def test_pkg_012_installed_with_wildcard_version(self):
        """
        This is a destructive test as it installs and then removes a package
        """
        if not self._WILDCARDS_SUPPORTED:
            self.skipTest("Wildcards in pkg.install are not supported")

        target = self._PKG_TARGETS[0]
        version = self.run_function("pkg.version", [target])

        # If this assert fails, we need to find new targets, this test needs to
        # be able to test successful installation of packages, so this package
        # needs to not be installed before we run the states below
        self.assertFalse(version)

        ret = self.run_state(
            "pkg.installed",
            name=target,
            version="*",
            refresh=False,
        )
        self.assertSaltTrueReturn(ret)

        # Repeat state, should pass
        ret = self.run_state(
            "pkg.installed",
            name=target,
            version="*",
            refresh=False,
        )

        expected_comment = (
            "All specified packages are already installed and are at the "
            "desired version"
        )
        self.assertSaltTrueReturn(ret)
        self.assertEqual(ret[next(iter(ret))]["comment"], expected_comment)

        # Repeat one more time with unavailable version, test should fail
        ret = self.run_state(
            "pkg.installed",
            name=target,
            version="93413*",
            refresh=False,
        )
        self.assertSaltFalseReturn(ret)

        # Clean up
        ret = self.run_state("pkg.removed", name=target)
        self.assertSaltTrueReturn(ret)

    @pytest.mark.requires_salt_modules("pkg.version", "pkg.latest_version")
    @pytest.mark.requires_salt_states("pkg.installed", "pkg.removed")
    @runs_on(kernel="linux", os_family=["Debian", "RedHat"])
    @pytest.mark.slow_test
    def test_pkg_013_installed_with_comparison_operator(self):
        """
        This is a destructive test as it installs and then removes a package
        """
        target = self._PKG_TARGETS[0]
        version = self.run_function("pkg.version", [target])

        # If this assert fails, we need to find new targets, this test needs to
        # be able to test successful installation of packages, so this package
        # needs to not be installed before we run the states below
        self.assertFalse(version)

        latest_version = self.run_function(
            "pkg.latest_version", [target], refresh=False
        )

        try:
            ret = self.run_state(
                "pkg.installed",
                name=target,
                version="<9999999",
                refresh=False,
            )
            self.assertSaltTrueReturn(ret)

            # The version that was installed should be the latest available
            version = self.run_function("pkg.version", [target])
            self.assertTrue(version, latest_version)
        finally:
            # Clean up
            ret = self.run_state("pkg.removed", name=target)
            self.assertSaltTrueReturn(ret)

    @pytest.mark.requires_salt_modules("pkg.version")
    @pytest.mark.requires_salt_states("pkg.installed", "pkg.removed")
    @runs_on(kernel="linux", os_familiy="RedHat")
    def test_pkg_014_installed_missing_release(self):
        """
        Tests that a version number missing the release portion still resolves
        as correctly installed. For example, version 2.0.2 instead of 2.0.2-1.el7
        """
        target = self._PKG_TARGETS[0]
        version = self.run_function("pkg.version", [target])

        # If this assert fails, we need to find new targets, this test needs to
        # be able to test successful installation of packages, so this package
        # needs to not be installed before we run the states below
        self.assertFalse(version)

        ret = self.run_state(
            "pkg.installed",
            name=target,
            version=salt.utils.pkg.rpm.version_to_evr(version)[1],
            refresh=False,
        )
        self.assertSaltTrueReturn(ret)

        # Clean up
        ret = self.run_state("pkg.removed", name=target)
        self.assertSaltTrueReturn(ret)

    @pytest.mark.requires_salt_modules(
        "pkg.hold", "pkg.unhold", "pkg.version", "pkg.list_pkgs"
    )
    @pytest.mark.requires_salt_states("pkg.installed", "pkg.removed")
    @requires_system_grains
    @pytest.mark.slow_test
    def test_pkg_015_installed_held(self, grains=None):
        """
        Tests that a package can be held even when the package is already installed.
        """
        versionlock_pkg = None
        if grains["os_family"] == "RedHat":
            pkgs = {
                p
                for p in self.run_function("pkg.list_repo_pkgs")
                if "yum-plugin-versionlock" in p
            }
            if not pkgs:
                self.skipTest("No versionlock package found in repositories")
            for versionlock_pkg in pkgs:
                ret = self.run_state(
                    "pkg.installed", name=versionlock_pkg, refresh=False
                )
                # Exit loop if a versionlock package installed correctly
                try:
                    self.assertSaltTrueReturn(ret)
                    log.debug("Installed versionlock package: %s", versionlock_pkg)
                    break
                except AssertionError as exc:
                    log.debug("Versionlock package not found:\n%s", exc)
            else:
                self.fail("Could not install versionlock package from {}".format(pkgs))

        target = self._PKG_TARGETS[0]

        # First we ensure that the package is installed
        ret = self.run_state(
            "pkg.installed",
            name=target,
            refresh=False,
        )
        self.assertSaltTrueReturn(ret)

        # Then we check that the package is now held
        ret = self.run_state(
            "pkg.installed",
            name=target,
            hold=True,
            refresh=False,
        )

        if versionlock_pkg and "-versionlock is not installed" in str(ret):
            self.skipTest("{}  `{}` is installed".format(ret, versionlock_pkg))

        # changes from pkg.hold for Red Hat family are different
        target_changes = {}
        if (
            grains["os_family"] == "RedHat"
            or grains["os"] == "FreeBSD"
            or grains["os_family"] == "Suse"
        ):
            target_changes = {"new": "hold", "old": ""}
        elif grains["os_family"] == "Debian":
            target_changes = {"new": "hold", "old": "install"}

        try:
            tag = "pkg_|-{0}_|-{0}_|-installed".format(target)
            self.assertSaltTrueReturn(ret)
            self.assertIn(tag, ret)
            self.assertIn("changes", ret[tag])
            self.assertIn(target, ret[tag]["changes"])
            if not target_changes:
                self.skipTest(
                    "Test needs to be configured for {}: {}".format(
                        grains["os"], ret[tag]["changes"][target]
                    )
                )
            self.assertEqual(ret[tag]["changes"][target], target_changes)
        finally:
            # Clean up, unhold package and remove
            self.run_function("pkg.unhold", name=target)
            ret = self.run_state("pkg.removed", name=target)
            self.assertSaltTrueReturn(ret)
            if versionlock_pkg:
                ret = self.run_state("pkg.removed", name=versionlock_pkg)
                self.assertSaltTrueReturn(ret)

    @pytest.mark.requires_salt_states("pkg.installed", "pkg.removed")
    def test_pkg_016_conditionally_ignore_epoch(self):
        """
        See
        https://github.com/saltstack/salt/issues/56654#issuecomment-615034952

        This is a destructive test as it installs a package
        """
        if not self._PKG_EPOCH_TARGETS:
            self.skipTest('No targets have been configured with "epoch" in the version')

        target = self._PKG_EPOCH_TARGETS[0]

        # Strip the epoch from the latest available version
        version = self.latest_version(target).split(":", 1)[-1]
        # If this assert fails, we need to find a new target. This test
        # needs to be able to test successful installation of the package, so
        # the target needs to not be installed before we run the
        # pkg.installed state below
        self.assertTrue(version)

        # CASE 1: package name passed in "name" param
        ret = self.run_state(
            "pkg.installed", name=target, version=version, refresh=False
        )
        self.assertSaltTrueReturn(ret)
        ret = self.run_state("pkg.removed", name=target)
        self.assertSaltTrueReturn(ret)

        # CASE 2: same as case 1 but with "pkgs"
        ret = self.run_state(
            "pkg.installed", name="foo", pkgs=[{target: version}], refresh=False
        )
        self.assertSaltTrueReturn(ret)
        ret = self.run_state("pkg.removed", name=target)
        self.assertSaltTrueReturn(ret)

    @pytest.mark.requires_salt_modules(
        "pkg.hold", "pkg.unhold", "pkg.version", "pkg.list_pkgs"
    )
    @pytest.mark.requires_salt_states("pkg.installed", "pkg.removed")
    @requires_system_grains
    @pytest.mark.slow_test
    def test_pkg_017_installed_held_equals_false(self, grains=None):
        """
        Tests that a package is installed when hold is explicitly False.

        See https://github.com/saltstack/salt/issues/58801.
        """
        versionlock_pkg = None
        if grains["os_family"] == "RedHat":
            from salt.modules.yumpkg import _versionlock_pkg

            pkgs = {
                p
                for p in self.run_function("pkg.list_repo_pkgs")
                if _versionlock_pkg(grains) in p
            }
            if not pkgs:
                self.skipTest("No versionlock package found in repositories")
            for versionlock_pkg in pkgs:
                ret = self.run_state(
                    "pkg.installed", name=versionlock_pkg, refresh=False
                )
                # Exit loop if a versionlock package installed correctly
                try:
                    self.assertSaltTrueReturn(ret)
                    log.debug("Installed versionlock package: %s", versionlock_pkg)
                    break
                except AssertionError as exc:
                    log.debug("Versionlock package not found:\n%s", exc)
            else:
                self.fail("Could not install versionlock package from {}".format(pkgs))

        target = self._PKG_TARGETS[0]

        # First we ensure that the package is installed
        target_ret = self.run_state(
            "pkg.installed",
            name=target,
            hold=False,
            refresh=False,
        )
        self.assertSaltTrueReturn(target_ret)

        if versionlock_pkg and "-versionlock is not installed" in str(target_ret):
            self.skipTest("{}  `{}` is installed".format(target_ret, versionlock_pkg))

        try:
            tag = "pkg_|-{0}_|-{0}_|-installed".format(target)
            self.assertSaltTrueReturn(target_ret)
            self.assertIn(tag, target_ret)
            self.assertIn("changes", target_ret[tag])
            # On Centos 7 package is already installed, no change happened
            if target_ret[tag].get("changes"):
                self.assertIn(target, target_ret[tag]["changes"])
            if grains["os_family"] == "Suse":
                self.assertIn("packages were installed", target_ret[tag]["comment"])
            else:
                #  The "held" string is part of a longer comment that may look
                #  like:
                #
                #    Package units is not being held.
                self.assertIn("held", target_ret[tag]["comment"])
        finally:
            # Clean up, unhold package and remove
            ret = self.run_state("pkg.removed", name=target)
            self.assertSaltTrueReturn(ret)
            if versionlock_pkg:
                ret = self.run_state("pkg.removed", name=versionlock_pkg)
                self.assertSaltTrueReturn(ret)

    @pytest.mark.requires_salt_modules("pkg.version")
    @pytest.mark.requires_salt_states("pkg.installed", "pkg.removed")
    def test_pkg_cap_001_installed(self):
        """
        This is a destructive test as it installs and then removes a package
        """
        if not self._PKG_CAP_TARGETS:
            self.skipTest("Capability not provided")

        target, realpkg = self._PKG_CAP_TARGETS[0]
        version = self.run_function("pkg.version", [target])
        realver = self.run_function("pkg.version", [realpkg])

        # If this condition is False, we need to find new targets.
        # This needs to be able to test successful installation of packages.
        # These packages need to not be installed before we run the states below
        if not (version and realver):
            self.skipTest("TODO: New pkg cap targets required")

        try:
            ret = self.run_state(
                "pkg.installed",
                name=target,
                refresh=False,
                resolve_capabilities=True,
                test=True,
            )
            self.assertInSaltComment(
                "The following packages would be installed/updated: {}".format(realpkg),
                ret,
            )
            ret = self.run_state(
                "pkg.installed", name=target, refresh=False, resolve_capabilities=True
            )
            self.assertSaltTrueReturn(ret)
        finally:
            ret = self.run_state("pkg.removed", name=realpkg)
            self.assertSaltTrueReturn(ret)

    @pytest.mark.requires_salt_states("pkg.installed", "pkg.removed")
    def test_pkg_cap_002_already_installed(self):
        """
        This is a destructive test as it installs and then removes a package
        """
        if not self._PKG_CAP_TARGETS:
            self.skipTest("Capability not provided")

        target, realpkg = self._PKG_CAP_TARGETS[0]
        version = self.run_function("pkg.version", [target])
        realver = self.run_function("pkg.version", [realpkg])

        # If this condition is False, we need to find new targets.
        # This needs to be able to test successful installation of packages.
        # These packages need to not be installed before we run the states below
        if not (version and realver):
            self.skipTest("TODO: New pkg cap targets required")

        try:
            # install the package
            ret = self.run_state("pkg.installed", name=realpkg, refresh=False)
            self.assertSaltTrueReturn(ret)

            # Try to install again. Nothing should be installed this time.
            ret = self.run_state(
                "pkg.installed",
                name=target,
                refresh=False,
                resolve_capabilities=True,
                test=True,
            )
            self.assertInSaltComment(
                "All specified packages are already installed", ret
            )

            ret = self.run_state(
                "pkg.installed", name=target, refresh=False, resolve_capabilities=True
            )
            self.assertSaltTrueReturn(ret)

            self.assertInSaltComment("packages are already installed", ret)
        finally:
            ret = self.run_state("pkg.removed", name=realpkg)
            self.assertSaltTrueReturn(ret)

    @skipIf(not _PKG_CAP_TARGETS, "Capability not available")
    @pytest.mark.requires_salt_states("pkg.installed", "pkg.removed")
    def test_pkg_cap_003_installed_multipkg_with_version(self):
        """
        This is a destructive test as it installs and then removes two packages
        """
        if not self._VERSION_SPEC_SUPPORTED:
            self.skipTest("Version specification not supported")
        target, realpkg = self._PKG_CAP_TARGETS[0]
        version = self.latest_version(target)
        realver = self.latest_version(realpkg)

        # If this condition is False, we need to find new targets.
        # This needs to be able to test successful installation of packages.
        # These packages need to not be installed before we run the states below
        if not (version and realver):
            self.skipTest("TODO: New pkg cap targets required")

        cleanup_pkgs = self._PKG_TARGETS
        try:
            pkgs = [
                {self._PKG_TARGETS[0]: version},
                self._PKG_TARGETS[1],
                {target: realver},
            ]
            ret = self.run_state(
                "pkg.installed",
                name="test_pkg_cap_003_installed_multipkg_with_version-install",
                pkgs=pkgs,
                refresh=False,
            )
            self.assertSaltFalseReturn(ret)

            ret = self.run_state(
                "pkg.installed",
                name="test_pkg_cap_003_installed_multipkg_with_version-install-capability",
                pkgs=pkgs,
                refresh=False,
                resolve_capabilities=True,
                test=True,
            )
            self.assertInSaltComment("packages would be installed/updated", ret)
            self.assertInSaltComment("{}={}".format(realpkg, realver), ret)

            ret = self.run_state(
                "pkg.installed",
                name="test_pkg_cap_003_installed_multipkg_with_version-install-capability",
                pkgs=pkgs,
                refresh=False,
                resolve_capabilities=True,
            )
            self.assertSaltTrueReturn(ret)
            cleanup_pkgs.append(realpkg)
        finally:
            ret = self.run_state(
                "pkg.removed",
                name="test_pkg_cap_003_installed_multipkg_with_version-remove",
                pkgs=cleanup_pkgs,
            )
            self.assertSaltTrueReturn(ret)

    @pytest.mark.requires_salt_modules("pkg.version")
    @pytest.mark.requires_salt_states("pkg.latest", "pkg.removed")
    def test_pkg_cap_004_latest(self):
        """
        This tests pkg.latest with a package that has no epoch (or a zero
        epoch).
        """
        if not self._PKG_CAP_TARGETS:
            self.skipTest("Capability not provided")

        target, realpkg = self._PKG_CAP_TARGETS[0]
        version = self.run_function("pkg.version", [target])
        realver = self.run_function("pkg.version", [realpkg])

        # If this condition is False, we need to find new targets.
        # This needs to be able to test successful installation of packages.
        # These packages need to not be installed before we run the states below
        if not (version and realver):
            self.skipTest("TODO: New pkg cap targets required")

        try:
            ret = self.run_state(
                "pkg.latest",
                name=target,
                refresh=False,
                resolve_capabilities=True,
                test=True,
            )
            self.assertInSaltComment(
                "The following packages would be installed/upgraded: {}".format(
                    realpkg
                ),
                ret,
            )
            ret = self.run_state(
                "pkg.latest", name=target, refresh=False, resolve_capabilities=True
            )
            self.assertSaltTrueReturn(ret)

            ret = self.run_state(
                "pkg.latest", name=target, refresh=False, resolve_capabilities=True
            )
            self.assertSaltTrueReturn(ret)
            self.assertInSaltComment("is already up-to-date", ret)
        finally:
            ret = self.run_state("pkg.removed", name=realpkg)
            self.assertSaltTrueReturn(ret)

    @pytest.mark.requires_salt_modules("pkg.version")
    @pytest.mark.requires_salt_states("pkg.installed", "pkg.removed", "pkg.downloaded")
    def test_pkg_cap_005_downloaded(self):
        """
        This is a destructive test as it installs and then removes a package
        """
        if not self._PKG_CAP_TARGETS:
            self.skipTest("Capability not provided")

        target, realpkg = self._PKG_CAP_TARGETS[0]
        version = self.run_function("pkg.version", [target])
        realver = self.run_function("pkg.version", [realpkg])

        # If this condition is False, we need to find new targets.
        # This needs to be able to test successful installation of packages.
        # These packages need to not be installed before we run the states below
        if not (version and realver):
            self.skipTest("TODO: New pkg cap targets required")

        ret = self.run_state("pkg.downloaded", name=target, refresh=False)
        self.assertSaltFalseReturn(ret)

        ret = self.run_state(
            "pkg.downloaded",
            name=target,
            refresh=False,
            resolve_capabilities=True,
            test=True,
        )
        self.assertInSaltComment(
            "The following packages would be downloaded: {}".format(realpkg), ret
        )

        ret = self.run_state(
            "pkg.downloaded", name=target, refresh=False, resolve_capabilities=True
        )
        self.assertSaltTrueReturn(ret)

    @pytest.mark.requires_salt_modules("pkg.version")
    @pytest.mark.requires_salt_states("pkg.installed", "pkg.removed", "pkg.uptodate")
    def test_pkg_cap_006_uptodate(self):
        """
        This is a destructive test as it installs and then removes a package
        """
        if not self._PKG_CAP_TARGETS:
            self.skipTest("Capability not provided")

        target, realpkg = self._PKG_CAP_TARGETS[0]
        version = self.run_function("pkg.version", [target])
        realver = self.run_function("pkg.version", [realpkg])

        # If this condition is False, we need to find new targets.
        # This needs to be able to test successful installation of packages.
        # These packages need to not be installed before we run the states below
        if not (version and realver):
            self.skipTest("TODO: New pkg cap targets required")

        try:
            ret = self.run_state(
                "pkg.installed", name=target, refresh=False, resolve_capabilities=True
            )
            self.assertSaltTrueReturn(ret)
            ret = self.run_state(
                "pkg.uptodate",
                name="test_pkg_cap_006_uptodate",
                pkgs=[target],
                refresh=False,
                resolve_capabilities=True,
            )
            self.assertSaltTrueReturn(ret)
            self.assertInSaltComment("System is already up-to-date", ret)
        finally:
            ret = self.run_state("pkg.removed", name=realpkg)
            self.assertSaltTrueReturn(ret)



#!/usr/bin/env python
#
# Copyright 2009 Facebook
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""HTTP utility code shared by clients and servers.

This module also defines the `HTTPServerRequest` class which is exposed
via `tornado.web.RequestHandler.request`.
"""
# pylint: skip-file

from __future__ import absolute_import, division, print_function

import calendar
import collections
import copy
import datetime
import email.utils
import numbers
import re
import time
from collections.abc import MutableMapping

from salt.ext.tornado.escape import native_str, parse_qs_bytes, utf8
from salt.ext.tornado.log import gen_log
from salt.ext.tornado.util import PY3, ObjectDict

if PY3:
    import http.cookies as Cookie
    from http.client import responses
    from urllib.parse import urlencode, urlparse, urlunparse, parse_qsl
else:
    import Cookie
    from httplib import responses
    from urllib import urlencode
    from urlparse import urlparse, urlunparse, parse_qsl


# responses is unused in this file, but we re-export it to other files.
# Reference it so pyflakes doesn't complain.
responses

try:
    from ssl import SSLError
except ImportError:
    # ssl is unavailable on app engine.
    class _SSLError(Exception):
        pass

    # Hack around a mypy limitation. We can't simply put "type: ignore"
    # on the class definition itself; must go through an assignment.
    SSLError = _SSLError  # type: ignore

try:
    import typing
except ImportError:
    pass


# RFC 7230 section 3.5: a recipient MAY recognize a single LF as a line
# terminator and ignore any preceding CR.
_CRLF_RE = re.compile(r"\r?\n")


class _NormalizedHeaderCache(dict):
    """Dynamic cached mapping of header names to Http-Header-Case.

    Implemented as a dict subclass so that cache hits are as fast as a
    normal dict lookup, without the overhead of a python function
    call.

    >>> normalized_headers = _NormalizedHeaderCache(10)
    >>> normalized_headers["coNtent-TYPE"]
    'Content-Type'
    """

    def __init__(self, size):
        super(_NormalizedHeaderCache, self).__init__()
        self.size = size
        self.queue = collections.deque()

    def __missing__(self, key):
        normalized = "-".join([w.capitalize() for w in key.split("-")])
        self[key] = normalized
        self.queue.append(key)
        if len(self.queue) > self.size:
            # Limit the size of the cache.  LRU would be better, but this
            # simpler approach should be fine.  In Python 2.7+ we could
            # use OrderedDict (or in 3.2+, @functools.lru_cache).
            old_key = self.queue.popleft()
            del self[old_key]
        return normalized


_normalized_headers = _NormalizedHeaderCache(1000)


class HTTPHeaders(MutableMapping):
    """A dictionary that maintains ``Http-Header-Case`` for all keys.

    Supports multiple values per key via a pair of new methods,
    `add()` and `get_list()`.  The regular dictionary interface
    returns a single value per key, with multiple values joined by a
    comma.

    >>> h = HTTPHeaders({"content-type": "text/html"})
    >>> list(h.keys())
    ['Content-Type']
    >>> h["Content-Type"]
    'text/html'

    >>> h.add("Set-Cookie", "A=B")
    >>> h.add("Set-Cookie", "C=D")
    >>> h["set-cookie"]
    'A=B,C=D'
    >>> h.get_list("set-cookie")
    ['A=B', 'C=D']

    >>> for (k,v) in sorted(h.get_all()):
    ...    print('%s: %s' % (k,v))
    ...
    Content-Type: text/html
    Set-Cookie: A=B
    Set-Cookie: C=D
    """

    def __init__(self, *args, **kwargs):
        self._dict = {}  # type: typing.Dict[str, str]
        self._as_list = {}  # type: typing.Dict[str, typing.List[str]]
        self._last_key = None
        if len(args) == 1 and len(kwargs) == 0 and isinstance(args[0], HTTPHeaders):
            # Copy constructor
            for k, v in args[0].get_all():
                self.add(k, v)
        else:
            # Dict-style initialization
            self.update(*args, **kwargs)

    # new public methods

    def add(self, name, value):
        # type: (str, str) -> None
        """Adds a new value for the given key."""
        norm_name = _normalized_headers[name]
        self._last_key = norm_name
        if norm_name in self:
            self._dict[norm_name] = (
                native_str(self[norm_name]) + "," + native_str(value)
            )
            self._as_list[norm_name].append(value)
        else:
            self[norm_name] = value

    def get_list(self, name):
        """Returns all values for the given header as a list."""
        norm_name = _normalized_headers[name]
        return self._as_list.get(norm_name, [])

    def get_all(self):
        # type: () -> typing.Iterable[typing.Tuple[str, str]]
        """Returns an iterable of all (name, value) pairs.

        If a header has multiple values, multiple pairs will be
        returned with the same name.
        """
        for name, values in self._as_list.items():
            for value in values:
                yield (name, value)

    def parse_line(self, line):
        """Updates the dictionary with a single header line.

        >>> h = HTTPHeaders()
        >>> h.parse_line("Content-Type: text/html")
        >>> h.get('content-type')
        'text/html'
        """
        if line[0].isspace():
            # continuation of a multi-line header
            new_part = " " + line.lstrip()
            self._as_list[self._last_key][-1] += new_part
            self._dict[self._last_key] += new_part
        else:
            name, value = line.split(":", 1)
            self.add(name, value.strip())

    @classmethod
    def parse(cls, headers):
        """Returns a dictionary from HTTP header text.

        >>> h = HTTPHeaders.parse("Content-Type: text/html\\r\\nContent-Length: 42\\r\\n")
        >>> sorted(h.items())
        [('Content-Length', '42'), ('Content-Type', 'text/html')]
        """
        h = cls()
        for line in _CRLF_RE.split(headers):
            if line:
                h.parse_line(line)
        return h

    # MutableMapping abstract method implementations.

    def __setitem__(self, name, value):
        norm_name = _normalized_headers[name]
        self._dict[norm_name] = value
        self._as_list[norm_name] = [value]

    def __getitem__(self, name):
        # type: (str) -> str
        return self._dict[_normalized_headers[name]]

    def __delitem__(self, name):
        norm_name = _normalized_headers[name]
        del self._dict[norm_name]
        del self._as_list[norm_name]

    def __len__(self):
        return len(self._dict)

    def __iter__(self):
        return iter(self._dict)

    def copy(self):
        # defined in dict but not in MutableMapping.
        return HTTPHeaders(self)

    # Use our overridden copy method for the copy.copy module.
    # This makes shallow copies one level deeper, but preserves
    # the appearance that HTTPHeaders is a single container.
    __copy__ = copy

    def __str__(self):
        lines = []
        for name, value in self.get_all():
            lines.append("%s: %s\n" % (name, value))
        return "".join(lines)

    __unicode__ = __str__


class HTTPServerRequest(object):
    """A single HTTP request.

    All attributes are type `str` unless otherwise noted.

    .. attribute:: method

       HTTP request method, e.g. "GET" or "POST"

    .. attribute:: uri

       The requested uri.

    .. attribute:: path

       The path portion of `uri`

    .. attribute:: query

       The query portion of `uri`

    .. attribute:: version

       HTTP version specified in request, e.g. "HTTP/1.1"

    .. attribute:: headers

       `.HTTPHeaders` dictionary-like object for request headers.  Acts like
       a case-insensitive dictionary with additional methods for repeated
       headers.

    .. attribute:: body

       Request body, if present, as a byte string.

    .. attribute:: remote_ip

       Client's IP address as a string.  If ``HTTPServer.xheaders`` is set,
       will pass along the real IP address provided by a load balancer
       in the ``X-Real-Ip`` or ``X-Forwarded-For`` header.

    .. versionchanged:: 3.1
       The list format of ``X-Forwarded-For`` is now supported.

    .. attribute:: protocol

       The protocol used, either "http" or "https".  If ``HTTPServer.xheaders``
       is set, will pass along the protocol used by a load balancer if
       reported via an ``X-Scheme`` header.

    .. attribute:: host

       The requested hostname, usually taken from the ``Host`` header.

    .. attribute:: arguments

       GET/POST arguments are available in the arguments property, which
       maps arguments names to lists of values (to support multiple values
       for individual names). Names are of type `str`, while arguments
       are byte strings.  Note that this is different from
       `.RequestHandler.get_argument`, which returns argument values as
       unicode strings.

    .. attribute:: query_arguments

       Same format as ``arguments``, but contains only arguments extracted
       from the query string.

       .. versionadded:: 3.2

    .. attribute:: body_arguments

       Same format as ``arguments``, but contains only arguments extracted
       from the request body.

       .. versionadded:: 3.2

    .. attribute:: files

       File uploads are available in the files property, which maps file
       names to lists of `.HTTPFile`.

    .. attribute:: connection

       An HTTP request is attached to a single HTTP connection, which can
       be accessed through the "connection" attribute. Since connections
       are typically kept open in HTTP/1.1, multiple requests can be handled
       sequentially on a single connection.

    .. versionchanged:: 4.0
       Moved from ``tornado.httpserver.HTTPRequest``.
    """

    def __init__(
        self,
        method=None,
        uri=None,
        version="HTTP/1.0",
        headers=None,
        body=None,
        host=None,
        files=None,
        connection=None,
        start_line=None,
        server_connection=None,
    ):
        if start_line is not None:
            method, uri, version = start_line
        self.method = method
        self.uri = uri
        self.version = version
        self.headers = headers or HTTPHeaders()
        self.body = body or b""

        # set remote IP and protocol
        context = getattr(connection, "context", None)
        self.remote_ip = getattr(context, "remote_ip", None)
        self.protocol = getattr(context, "protocol", "http")

        self.host = host or self.headers.get("Host") or "127.0.0.1"
        self.host_name = split_host_and_port(self.host.lower())[0]
        self.files = files or {}
        self.connection = connection
        self.server_connection = server_connection
        self._start_time = time.time()
        self._finish_time = None

        self.path, sep, self.query = uri.partition("?")
        self.arguments = parse_qs_bytes(self.query, keep_blank_values=True)
        self.query_arguments = copy.deepcopy(self.arguments)
        self.body_arguments = {}

    def supports_http_1_1(self):
        """Returns True if this request supports HTTP/1.1 semantics.

        .. deprecated:: 4.0
           Applications are less likely to need this information with the
           introduction of `.HTTPConnection`.  If you still need it, access
           the ``version`` attribute directly.
        """
        return self.version == "HTTP/1.1"

    @property
    def cookies(self):
        """A dictionary of Cookie.Morsel objects."""
        if not hasattr(self, "_cookies"):
            self._cookies = Cookie.SimpleCookie()
            if "Cookie" in self.headers:
                try:
                    parsed = parse_cookie(self.headers["Cookie"])
                except Exception:
                    pass
                else:
                    for k, v in parsed.items():
                        try:
                            self._cookies[k] = v
                        except Exception:
                            # SimpleCookie imposes some restrictions on keys;
                            # parse_cookie does not. Discard any cookies
                            # with disallowed keys.
                            pass
        return self._cookies

    def write(self, chunk, callback=None):
        """Writes the given chunk to the response stream.

        .. deprecated:: 4.0
           Use ``request.connection`` and the `.HTTPConnection` methods
           to write the response.
        """
        assert isinstance(chunk, bytes)
        assert self.version.startswith(
            "HTTP/1."
        ), "deprecated interface only supported in HTTP/1.x"
        self.connection.write(chunk, callback=callback)

    def finish(self):
        """Finishes this HTTP request on the open connection.

        .. deprecated:: 4.0
           Use ``request.connection`` and the `.HTTPConnection` methods
           to write the response.
        """
        self.connection.finish()
        self._finish_time = time.time()

    def full_url(self):
        """Reconstructs the full URL for this request."""
        return self.protocol + "://" + self.host + self.uri

    def request_time(self):
        """Returns the amount of time it took for this request to execute."""
        if self._finish_time is None:
            return time.time() - self._start_time
        else:
            return self._finish_time - self._start_time

    def get_ssl_certificate(self, binary_form=False):
        """Returns the client's SSL certificate, if any.

        To use client certificates, the HTTPServer's
        `ssl.SSLContext.verify_mode` field must be set, e.g.::

            ssl_ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
            ssl_ctx.load_cert_chain("foo.crt", "foo.key")
            ssl_ctx.load_verify_locations("cacerts.pem")
            ssl_ctx.verify_mode = ssl.CERT_REQUIRED
            server = HTTPServer(app, ssl_options=ssl_ctx)

        By default, the return value is a dictionary (or None, if no
        client certificate is present).  If ``binary_form`` is true, a
        DER-encoded form of the certificate is returned instead.  See
        SSLSocket.getpeercert() in the standard library for more
        details.
        http://docs.python.org/library/ssl.html#sslsocket-objects
        """
        try:
            return self.connection.stream.socket.getpeercert(binary_form=binary_form)
        except SSLError:
            return None

    def _parse_body(self):
        parse_body_arguments(
            self.headers.get("Content-Type", ""),
            self.body,
            self.body_arguments,
            self.files,
            self.headers,
        )

        for k, v in self.body_arguments.items():
            self.arguments.setdefault(k, []).extend(v)

    def __repr__(self):
        attrs = ("protocol", "host", "method", "uri", "version", "remote_ip")
        args = ", ".join(["%s=%r" % (n, getattr(self, n)) for n in attrs])
        return "%s(%s, headers=%s)" % (
            self.__class__.__name__,
            args,
            dict(self.headers),
        )


class HTTPInputError(Exception):
    """Exception class for malformed HTTP requests or responses
    from remote sources.

    .. versionadded:: 4.0
    """

    pass


class HTTPOutputError(Exception):
    """Exception class for errors in HTTP output.

    .. versionadded:: 4.0
    """

    pass


class HTTPServerConnectionDelegate(object):
    """Implement this interface to handle requests from `.HTTPServer`.

    .. versionadded:: 4.0
    """

    def start_request(self, server_conn, request_conn):
        """This method is called by the server when a new request has started.

        :arg server_conn: is an opaque object representing the long-lived
            (e.g. tcp-level) connection.
        :arg request_conn: is a `.HTTPConnection` object for a single
            request/response exchange.

        This method should return a `.HTTPMessageDelegate`.
        """
        raise NotImplementedError()

    def on_close(self, server_conn):
        """This method is called when a connection has been closed.

        :arg server_conn: is a server connection that has previously been
            passed to ``start_request``.
        """
        pass


class HTTPMessageDelegate(object):
    """Implement this interface to handle an HTTP request or response.

    .. versionadded:: 4.0
    """

    def headers_received(self, start_line, headers):
        """Called when the HTTP headers have been received and parsed.

        :arg start_line: a `.RequestStartLine` or `.ResponseStartLine`
            depending on whether this is a client or server message.
        :arg headers: a `.HTTPHeaders` instance.

        Some `.HTTPConnection` methods can only be called during
        ``headers_received``.

        May return a `.Future`; if it does the body will not be read
        until it is done.
        """
        pass

    def data_received(self, chunk):
        """Called when a chunk of data has been received.

        May return a `.Future` for flow control.
        """
        pass

    def finish(self):
        """Called after the last chunk of data has been received."""
        pass

    def on_connection_close(self):
        """Called if the connection is closed without finishing the request.

        If ``headers_received`` is called, either ``finish`` or
        ``on_connection_close`` will be called, but not both.
        """
        pass


class HTTPConnection(object):
    """Applications use this interface to write their responses.

    .. versionadded:: 4.0
    """

    def write_headers(self, start_line, headers, chunk=None, callback=None):
        """Write an HTTP header block.

        :arg start_line: a `.RequestStartLine` or `.ResponseStartLine`.
        :arg headers: a `.HTTPHeaders` instance.
        :arg chunk: the first (optional) chunk of data.  This is an optimization
            so that small responses can be written in the same call as their
            headers.
        :arg callback: a callback to be run when the write is complete.

        The ``version`` field of ``start_line`` is ignored.

        Returns a `.Future` if no callback is given.
        """
        raise NotImplementedError()

    def write(self, chunk, callback=None):
        """Writes a chunk of body data.

        The callback will be run when the write is complete.  If no callback
        is given, returns a Future.
        """
        raise NotImplementedError()

    def finish(self):
        """Indicates that the last body data has been written.
        """
        raise NotImplementedError()


def url_concat(url, args):
    """Concatenate url and arguments regardless of whether
    url has existing query parameters.

    ``args`` may be either a dictionary or a list of key-value pairs
    (the latter allows for multiple values with the same key.

    >>> url_concat("http://example.com/foo", dict(c="d"))
    'http://example.com/foo?c=d'
    >>> url_concat("http://example.com/foo?a=b", dict(c="d"))
    'http://example.com/foo?a=b&c=d'
    >>> url_concat("http://example.com/foo?a=b", [("c", "d"), ("c", "d2")])
    'http://example.com/foo?a=b&c=d&c=d2'
    """
    if args is None:
        return url
    parsed_url = urlparse(url)
    if isinstance(args, dict):
        parsed_query = parse_qsl(parsed_url.query, keep_blank_values=True)
        parsed_query.extend(args.items())
    elif isinstance(args, list) or isinstance(args, tuple):
        parsed_query = parse_qsl(parsed_url.query, keep_blank_values=True)
        parsed_query.extend(args)
    else:
        err = "'args' parameter should be dict, list or tuple. Not {0}".format(
            type(args)
        )
        raise TypeError(err)
    final_query = urlencode(parsed_query)
    url = urlunparse(
        (
            parsed_url[0],
            parsed_url[1],
            parsed_url[2],
            parsed_url[3],
            final_query,
            parsed_url[5],
        )
    )
    return url


class HTTPFile(ObjectDict):
    """Represents a file uploaded via a form.

    For backwards compatibility, its instance attributes are also
    accessible as dictionary keys.

    * ``filename``
    * ``body``
    * ``content_type``
    """

    pass


def _parse_request_range(range_header):
    """Parses a Range header.

    Returns either ``None`` or tuple ``(start, end)``.
    Note that while the HTTP headers use inclusive byte positions,
    this method returns indexes suitable for use in slices.

    >>> start, end = _parse_request_range("bytes=1-2")
    >>> start, end
    (1, 3)
    >>> [0, 1, 2, 3, 4][start:end]
    [1, 2]
    >>> _parse_request_range("bytes=6-")
    (6, None)
    >>> _parse_request_range("bytes=-6")
    (-6, None)
    >>> _parse_request_range("bytes=-0")
    (None, 0)
    >>> _parse_request_range("bytes=")
    (None, None)
    >>> _parse_request_range("foo=42")
    >>> _parse_request_range("bytes=1-2,6-10")

    Note: only supports one range (ex, ``bytes=1-2,6-10`` is not allowed).

    See [0] for the details of the range header.

    [0]: http://greenbytes.de/tech/webdav/draft-ietf-httpbis-p5-range-latest.html#byte.ranges
    """
    unit, _, value = range_header.partition("=")
    unit, value = unit.strip(), value.strip()
    if unit != "bytes":
        return None
    start_b, _, end_b = value.partition("-")
    try:
        start = _int_or_none(start_b)
        end = _int_or_none(end_b)
    except ValueError:
        return None
    if end is not None:
        if start is None:
            if end != 0:
                start = -end
                end = None
        else:
            end += 1
    return (start, end)


def _get_content_range(start, end, total):
    """Returns a suitable Content-Range header:

    >>> print(_get_content_range(None, 1, 4))
    bytes 0-0/4
    >>> print(_get_content_range(1, 3, 4))
    bytes 1-2/4
    >>> print(_get_content_range(None, None, 4))
    bytes 0-3/4
    """
    start = start or 0
    end = (end or total) - 1
    return "bytes %s-%s/%s" % (start, end, total)


def _int_or_none(val):
    val = val.strip()
    if val == "":
        return None
    return int(val)


def parse_body_arguments(content_type, body, arguments, files, headers=None):
    """Parses a form request body.

    Supports ``application/x-www-form-urlencoded`` and
    ``multipart/form-data``.  The ``content_type`` parameter should be
    a string and ``body`` should be a byte string.  The ``arguments``
    and ``files`` parameters are dictionaries that will be updated
    with the parsed contents.
    """
    if headers and "Content-Encoding" in headers:
        gen_log.warning("Unsupported Content-Encoding: %s", headers["Content-Encoding"])
        return
    if content_type.startswith("application/x-www-form-urlencoded"):
        try:
            uri_arguments = parse_qs_bytes(native_str(body), keep_blank_values=True)
        except Exception as e:
            gen_log.warning("Invalid x-www-form-urlencoded body: %s", e)
            uri_arguments = {}
        for name, values in uri_arguments.items():
            if values:
                arguments.setdefault(name, []).extend(values)
    elif content_type.startswith("multipart/form-data"):
        try:
            fields = content_type.split(";")
            for field in fields:
                k, sep, v = field.strip().partition("=")
                if k == "boundary" and v:
                    parse_multipart_form_data(utf8(v), body, arguments, files)
                    break
            else:
                raise ValueError("multipart boundary not found")
        except Exception as e:
            gen_log.warning("Invalid multipart/form-data: %s", e)


def parse_multipart_form_data(boundary, data, arguments, files):
    """Parses a ``multipart/form-data`` body.

    The ``boundary`` and ``data`` parameters are both byte strings.
    The dictionaries given in the arguments and files parameters
    will be updated with the contents of the body.
    """
    # The standard allows for the boundary to be quoted in the header,
    # although it's rare (it happens at least for google app engine
    # xmpp).  I think we're also supposed to handle backslash-escapes
    # here but I'll save that until we see a client that uses them
    # in the wild.
    if boundary.startswith(b'"') and boundary.endswith(b'"'):
        boundary = boundary[1:-1]
    final_boundary_index = data.rfind(b"--" + boundary + b"--")
    if final_boundary_index == -1:
        gen_log.warning("Invalid multipart/form-data: no final boundary")
        return
    parts = data[:final_boundary_index].split(b"--" + boundary + b"\r\n")
    for part in parts:
        if not part:
            continue
        eoh = part.find(b"\r\n\r\n")
        if eoh == -1:
            gen_log.warning("multipart/form-data missing headers")
            continue
        headers = HTTPHeaders.parse(part[:eoh].decode("utf-8"))
        disp_header = headers.get("Content-Disposition", "")
        disposition, disp_params = _parse_header(disp_header)
        if disposition != "form-data" or not part.endswith(b"\r\n"):
            gen_log.warning("Invalid multipart/form-data")
            continue
        value = part[eoh + 4 : -2]
        if not disp_params.get("name"):
            gen_log.warning("multipart/form-data value missing name")
            continue
        name = disp_params["name"]
        if disp_params.get("filename"):
            ctype = headers.get("Content-Type", "application/unknown")
            files.setdefault(name, []).append(
                HTTPFile(  # type: ignore
                    filename=disp_params["filename"], body=value, content_type=ctype
                )
            )
        else:
            arguments.setdefault(name, []).append(value)


def format_timestamp(ts):
    """Formats a timestamp in the format used by HTTP.

    The argument may be a numeric timestamp as returned by `time.time`,
    a time tuple as returned by `time.gmtime`, or a `datetime.datetime`
    object.

    >>> format_timestamp(1359312200)
    'Sun, 27 Jan 2013 18:43:20 GMT'
    """
    if isinstance(ts, numbers.Real):
        pass
    elif isinstance(ts, (tuple, time.struct_time)):
        ts = calendar.timegm(ts)
    elif isinstance(ts, datetime.datetime):
        ts = calendar.timegm(ts.utctimetuple())
    else:
        raise TypeError("unknown timestamp type: %r" % ts)
    return email.utils.formatdate(ts, usegmt=True)


RequestStartLine = collections.namedtuple(
    "RequestStartLine", ["method", "path", "version"]
)


def parse_request_start_line(line):
    """Returns a (method, path, version) tuple for an HTTP 1.x request line.

    The response is a `collections.namedtuple`.

    >>> parse_request_start_line("GET /foo HTTP/1.1")
    RequestStartLine(method='GET', path='/foo', version='HTTP/1.1')
    """
    try:
        method, path, version = line.split(" ")
    except ValueError:
        raise HTTPInputError("Malformed HTTP request line")
    if not re.match(r"^HTTP/1\.[0-9]$", version):
        raise HTTPInputError(
            "Malformed HTTP version in HTTP Request-Line: %r" % version
        )
    return RequestStartLine(method, path, version)


ResponseStartLine = collections.namedtuple(
    "ResponseStartLine", ["version", "code", "reason"]
)


def parse_response_start_line(line):
    """Returns a (version, code, reason) tuple for an HTTP 1.x response line.

    The response is a `collections.namedtuple`.

    >>> parse_response_start_line("HTTP/1.1 200 OK")
    ResponseStartLine(version='HTTP/1.1', code=200, reason='OK')
    """
    line = native_str(line)
    match = re.match("(HTTP/1.[0-9]) ([0-9]+) ([^\r]*)", line)
    if not match:
        raise HTTPInputError("Error parsing response start line")
    return ResponseStartLine(match.group(1), int(match.group(2)), match.group(3))


# _parseparam and _parse_header are copied and modified from python2.7's cgi.py
# The original 2.7 version of this code did not correctly support some
# combinations of semicolons and double quotes.
# It has also been modified to support valueless parameters as seen in
# websocket extension negotiations.


def _parseparam(s):
    while s[:1] == ";":
        s = s[1:]
        end = s.find(";")
        while end > 0 and (s.count('"', 0, end) - s.count('\\"', 0, end)) % 2:
            end = s.find(";", end + 1)
        if end < 0:
            end = len(s)
        f = s[:end]
        yield f.strip()
        s = s[end:]


def _parse_header(line):
    """Parse a Content-type like header.

    Return the main content-type and a dictionary of options.

    """
    parts = _parseparam(";" + line)
    key = next(parts)
    pdict = {}
    for p in parts:
        i = p.find("=")
        if i >= 0:
            name = p[:i].strip().lower()
            value = p[i + 1 :].strip()
            if len(value) >= 2 and value[0] == value[-1] == '"':
                value = value[1:-1]
                value = value.replace("\\\\", "\\").replace('\\"', '"')
            pdict[name] = value
        else:
            pdict[p] = None
    return key, pdict


def _encode_header(key, pdict):
    """Inverse of _parse_header.

    >>> _encode_header('permessage-deflate',
    ...     {'client_max_window_bits': 15, 'client_no_context_takeover': None})
    'permessage-deflate; client_max_window_bits=15; client_no_context_takeover'
    """
    if not pdict:
        return key
    out = [key]
    # Sort the parameters just to make it easy to test.
    for k, v in sorted(pdict.items()):
        if v is None:
            out.append(k)
        else:
            # TODO: quote if necessary.
            out.append("%s=%s" % (k, v))
    return "; ".join(out)


def doctests():
    import doctest

    return doctest.DocTestSuite()


def split_host_and_port(netloc):
    """Returns ``(host, port)`` tuple from ``netloc``.

    Returned ``port`` will be ``None`` if not present.

    .. versionadded:: 4.1
    """
    match = re.match(r"^(.+):(\d+)$", netloc)
    if match:
        host = match.group(1)
        port = int(match.group(2))
    else:
        host = netloc
        port = None
    return (host, port)


_OctalPatt = re.compile(r"\\[0-3][0-7][0-7]")
_QuotePatt = re.compile(r"[\\].")
_nulljoin = "".join


def _unquote_cookie(str):
    """Handle double quotes and escaping in cookie values.

    This method is copied verbatim from the Python 3.5 standard
    library (http.cookies._unquote) so we don't have to depend on
    non-public interfaces.
    """
    # If there aren't any doublequotes,
    # then there can't be any special characters.  See RFC 2109.
    if str is None or len(str) < 2:
        return str
    if str[0] != '"' or str[-1] != '"':
        return str

    # We have to assume that we must decode this string.
    # Down to work.

    # Remove the "s
    str = str[1:-1]

    # Check for special sequences.  Examples:
    #    \012 --> \n
    #    \"   --> "
    #
    i = 0
    n = len(str)
    res = []
    while 0 <= i < n:
        o_match = _OctalPatt.search(str, i)
        q_match = _QuotePatt.search(str, i)
        if not o_match and not q_match:  # Neither matched
            res.append(str[i:])
            break
        # else:
        j = k = -1
        if o_match:
            j = o_match.start(0)
        if q_match:
            k = q_match.start(0)
        if q_match and (not o_match or k < j):  # QuotePatt matched
            res.append(str[i:k])
            res.append(str[k + 1])
            i = k + 2
        else:  # OctalPatt matched
            res.append(str[i:j])
            res.append(chr(int(str[j + 1 : j + 4], 8)))
            i = j + 4
    return _nulljoin(res)


def parse_cookie(cookie):
    """Parse a ``Cookie`` HTTP header into a dict of name/value pairs.

    This function attempts to mimic browser cookie parsing behavior;
    it specifically does not follow any of the cookie-related RFCs
    (because browsers don't either).

    The algorithm used is identical to that used by Django version 1.9.10.

    .. versionadded:: 4.4.2
    """
    cookiedict = {}
    for chunk in cookie.split(str(";")):
        if str("=") in chunk:
            key, val = chunk.split(str("="), 1)
        else:
            # Assume an empty name per
            # https://bugzilla.mozilla.org/show_bug.cgi?id=169091
            key, val = str(""), chunk
        key, val = key.strip(), val.strip()
        if key or val:
            # unquote using Python's algorithm.
            cookiedict[key] = _unquote_cookie(val)
    return cookiedict
